# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python [conda env:openpiv] *
#     language: python
#     name: conda-env-openpiv-py
# ---

# %% [markdown]
# # OpenPIV on the bridgepile_wake
#
# See the post on LinkedIn by Stefano Brizzolara
# https://www.linkedin.com/posts/stefano-brizzolara-6a8501198_rheinfall-flowvisualization-ugcPost-6672832128742408192-lRub

# %%
from openpiv import tools, process, validation, filters, scaling 

import numpy as np
import matplotlib.pyplot as plt
# %matplotlib inline

import imageio


# %%
frame_a  = tools.imread( '~/Downloads/bridgepile_wake/frame0001.tif' )
frame_b  = tools.imread( '~/Downloads/bridgepile_wake/frame0011.tif' )

# %%
fig,ax = plt.subplots(1,2,figsize=(12,10))
ax[0].imshow(frame_a,cmap=plt.cm.gray)
ax[1].imshow(frame_b,cmap=plt.cm.gray)


# %%
winsize = 48 # pixels
searchsize = 96  # pixels, search in image B
overlap = 24 # pixels
dt = 1./30 # sec, assume 30 fps


u0, v0, sig2noise = process.extended_search_area_piv( frame_a.astype(np.int32), frame_b.astype(np.int32), window_size=winsize, overlap=overlap, dt=dt, search_area_size=searchsize, sig2noise_method='peak2peak' )

# %%
x, y = process.get_coordinates( image_size=frame_a.shape, window_size=winsize, overlap=overlap )

# %%
u1, v1, mask = validation.sig2noise_val( u0, v0, sig2noise, threshold = 1.15)

# %%
u2, v2 = filters.replace_outliers( u1, v1, method='localmean', max_iter=10, kernel_size=2)

# %%
x, y, u3, v3 = scaling.uniform(x, y, u2, v2, scaling_factor = 1. )

# %%
tools.save(x, y, u3, v3, mask, 'exp1_001.txt' )

# %%
# tools.display_vector_field('exp1_001.txt', scaling_factor=100., width=0.0025)

# %%
# If you need a larger view:
fig, ax = plt.subplots(figsize=(12,12))
tools.display_vector_field('exp1_001.txt', ax=ax, scaling_factor=1.0, scale=3500, width=0.0045, on_img=True, image_name='~/Downloads/bridgepile_wake/frame0001.tif');

# %%

# %%



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %%
#Run the new window deformation

# %%
from openpiv import windef
from openpiv import tools, process, scaling, validation, filters, preprocess
import numpy as np
import os
from time import time
import warnings


import matplotlib.pyplot as plt
# %matplotlib inline  

# %%
settings = windef.Settings()


'Data related settings'
# Folder with the images to process
settings.filepath_images = '../test1/'
# Folder for the outputs
settings.save_path = '../test1/'
# Root name of the output Folder for Result Files
settings.save_folder_suffix = 'Test_1'
# Format and Image Sequence
settings.frame_pattern_a = 'exp1_001_a.bmp'
settings.frame_pattern_b = 'exp1_001_b.bmp'

'Region of interest'
# (50,300,50,300) #Region of interest: (xmin,xmax,ymin,ymax) or 'full' for full image
settings.ROI = 'full'

'Image preprocessing'
# 'None' for no masking, 'edges' for edges masking, 'intensity' for intensity masking
# WARNING: This part is under development so better not to use MASKS
settings.dynamic_masking_method = 'None'
settings.dynamic_masking_threshold = 0.005
settings.dynamic_masking_filter_size = 7

'Processing Parameters'
settings.correlation_method='circular'  # 'circular' or 'linear'
settings.iterations = 2  # select the number of PIV passes
# add the interroagtion window size for each pass. 
# For the moment, it should be a power of 2 
settings.windowsizes = (64, 32, 16) # if longer than n iteration the rest is ignored
# The overlap of the interroagtion window for each pass.
settings.overlap = (32, 16, 8) # This is 50% overlap
# Has to be a value with base two. In general window size/2 is a good choice.
# methode used for subpixel interpolation: 'gaussian','centroid','parabolic'
settings.subpixel_method = 'gaussian'
# order of the image interpolation for the window deformation
settings.interpolation_order = 3
settings.scaling_factor = 1  # scaling factor pixel/meter
settings.dt = 1  # time between to frames (in seconds)
'Signal to noise ratio options (only for the last pass)'
# It is possible to decide if the S/N should be computed (for the last pass) or not
settings.extract_sig2noise = True  # 'True' or 'False' (only for the last pass)
# method used to calculate the signal to noise ratio 'peak2peak' or 'peak2mean'
settings.sig2noise_method = 'peak2peak'
# select the width of the masked to masked out pixels next to the main peak
settings.sig2noise_mask = 2
# If extract_sig2noise==False the values in the signal to noise ratio
# output column are set to NaN
'vector validation options'
# choose if you want to do validation of the first pass: True or False
settings.validation_first_pass = True
# only effecting the first pass of the interrogation the following passes
# in the multipass will be validated
'Validation Parameters'
# The validation is done at each iteration based on three filters.
# The first filter is based on the min/max ranges. Observe that these values are defined in
# terms of minimum and maximum displacement in pixel/frames.
settings.MinMax_U_disp = (-30, 30)
settings.MinMax_V_disp = (-30, 30)
# The second filter is based on the global STD threshold
settings.std_threshold = 7  # threshold of the std validation
# The third filter is the median test (not normalized at the moment)
settings.median_threshold = 3  # threshold of the median validation
# On the last iteration, an additional validation can be done based on the S/N.
settings.median_size=1 #defines the size of the local median
'Validation based on the signal to noise ratio'
# Note: only available when extract_sig2noise==True and only for the last
# pass of the interrogation
# Enable the signal to noise ratio validation. Options: True or False
settings.do_sig2noise_validation = False # This is time consuming
# minmum signal to noise ratio that is need for a valid vector
settings.sig2noise_threshold = 1.2
'Outlier replacement or Smoothing options'
# Replacment options for vectors which are masked as invalid by the validation
settings.replace_vectors = True # Enable the replacment. Chosse: True or False
settings.smoothn=True #Enables smoothing of the displacemenet field
settings.smoothn_p=0.5 # This is a smoothing parameter
# select a method to replace the outliers: 'localmean', 'disk', 'distance'
settings.filter_method = 'localmean'
# maximum iterations performed to replace the outliers
settings.max_filter_iteration = 4
settings.filter_kernel_size = 2  # kernel size for the localmean method
'Output options'
# Select if you want to save the plotted vectorfield: True or False
settings.save_plot = False
# Choose wether you want to see the vectorfield or not :True or False
settings.show_plot = True
settings.scale_plot = 200  # select a value to scale the quiver plot of the vectorfield
# run the script with the given settings

# %%
dir(windef)

# %%
windef.piv(settings)

# %%
#Run the extended search area PIV

# %%


# we can run it from any folder
path = settings.filepath_images


frame_a  = tools.imread( os.path.join(path,settings.frame_pattern_a))
frame_b  = tools.imread( os.path.join(path,settings.frame_pattern_b))

frame_a = (frame_a).astype(np.int32)
frame_b = (frame_b).astype(np.int32)

u, v, sig2noise = process.extended_search_area_piv( frame_a, frame_b, \
    window_size=32, overlap=16, dt=1, search_area_size=64, sig2noise_method='peak2peak' )
x, y = process.get_coordinates( image_size=frame_a.shape, window_size=32, overlap=16 )
u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 1.3 )
u, v, mask = validation.global_val( u, v, (-1000, 2000), (-1000, 1000) )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 1)
tools.save(x, y, u, v, mask, 'test1.vec' )
tools.display_vector_field('test1.vec', scale=75, width=0.0035)

# %%
#Run the widim

# %%


scaling_factor = 1

# we can run it from any folder
path = settings.filepath_images


frame_a  = tools.imread( os.path.join(path,settings.frame_pattern_a))
frame_b  = tools.imread( os.path.join(path,settings.frame_pattern_b))

#no background removal will be performed so 'mark' is initialized to 1 everywhere
mark = np.zeros(frame_a.shape, dtype=np.int32)
for I in range(mark.shape[0]):
    for J in range(mark.shape[1]):
        mark[I,J]=1

#main algorithm
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    x,y,u,v, mask=process.WiDIM( frame_a.astype(np.int32), frame_b.astype(np.int32), mark, min_window_size=16, overlap_ratio=0.0, coarse_factor=2, dt=1, validation_method='mean_velocity', trust_1st_iter=1, validation_iter=1, tolerance=0.7, nb_iter_max=3, sig2noise_method='peak2peak')

#display results
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = scaling_factor )

tools.save(x, y, u, v, mask, '2image_00.txt' )

tools.display_vector_field('2image_00.txt',widim=True,on_img=False, image_name=os.path.join(path,'../test2/2image_00.tif'), window_size=16, scaling_factor=scaling_factor, scale=200, width=0.001)

#further validation can be performed to eliminate the few remaining wrong vectors


# %%



# Compilation of tools to do post processing of data,
# including calculations of vorticity and temporal analysis such as calculation of mean and fluctuating velocity components, turbulent kinetic energy and Reynolds stresses
# Vorticity calculations are accurate! it has been validated against base Matlab code and Tecplot calculaions. All calculations are performed using vector calculus to reduce run time.
# By: Pouya Mohtat
# Revision 0.2: Nov, 2020

# To do:
# ------------
# 1- A second order difference method can be implimented to calculate vorticity. It may increase the robustness of the vorticity results to noisy velocity signals.

# 2- This script takes the unvalidated piv outputs and does post processing on those results. in the GUI we are doing the same thing only that we are doing it on 
#    x, y, u, v results directly instead of reading it from files (to increase speed, obviously). so there is some duplicate code that could be refracted.


import numpy as np
from openpivtk import tools, validation, smoothn, filters, scaling
import matplotlib.pyplot as plt
import os, glob
import multiprocessing
from functools import partial
from collections import OrderedDict
from configparser import ConfigParser


def vorticity(u,v,x,y):
    """"Calculates vorticity for a velocity field with coordinates x and y
    the vorticity field is calculated using a first order central difference method

    Parameters
    ----------
    u, v : 2d np.ndarray
        u and v velocity fields

    x, y : 2d np.ndarray
        two dimentional arrays containing x and y coordinates of the mesh grid 

    Returns
    -------
    vor : 2d np.ndarray
        the vorticity field
    """
    dv_dx = np.zeros(x.shape)
    du_dy = np.zeros(y.shape)
    nx = np.unique(x)
    ny = np.unique(y)
    for iy in range(len(ny)):
        dv_dx[iy,:] = getSlope( v[iy,:], x[iy,:] )
    for ix in range(len(nx)):
        du_dy[:,ix] = getSlope( u[:,ix], y[:,ix] )
    return dv_dx - du_dy

def getSlope(y,x):
    """calculates the dy/dx slope: (y(n+1) - y(n)) / (x(n+1) - x(n))

    Parameters
    ----------
    y,x : 1d np.ndarray 
        1 dimentional arrays containing x and y(x) values

    Returns
    -------
    slope : 1d np.ndarray
        the dy/dx array
    """
    xp = np.pad(x, 1, 'edge')
    yp = np.pad(y, 1, 'edge')
    slope = (yp[2:] - yp[:-2])/(xp[2:] - xp[:-2])
    return slope

def calcAvgField(file_list, col_list=[2,3]):
    """Calculates the average field for the given list of columns

    Parameters
    ----------
    file_list : List of strings
        absolute path to the files containing flow data

    col_list : list of columns
        a list of columns for which the average needs to be calculated.
        the default is columns 2 and 3 which represent the u and v velocities respectively.

    Returns
    -------
    AVG : tuple of n arrays
        the average field data. the number of elements in the returned tuple depends on the len(col_list)
    """
    ntime = len(file_list)
    sample = np.loadtxt(file_list[0], skiprows=1)
    AVG = np.zeros((sample.shape[0], len(col_list)), sample.dtype)
    for i in range(ntime):
        data = np.loadtxt(file_list[i], skiprows=1)
        AVG += data[:,col_list]
    AVG = AVG / ntime

    return np.hsplit(AVG, AVG.shape[1])


def postProcess(file_name, pro, pos, saveLoc=None, header=None):
    """Carries out the post processing for a minimall output. does the following:
        validation and vector replacement, smoothing, field manipulation, scaling 
        and calculation of vorticity and velocity magnitude


    Parameters
    ----------
    file_name : string
        absolute path to the data file

    pro, pos : dictionary
        dictionaries containing processing and postprocessing settings

    saveLoc: string
        absolute path to the folder for saving results. default is None which means
        results are returned without saving to file.

    header: string
        the header line that is saved to the file.

    Returns
    -------
    basename : string
        file name
    
    x, y, u, v, mask, vor, velMag : 2D np.ndarray
        field data
    """
    #read file
    x, y, u, v, s2n = tools.read_data(file_name)
    #validation
    mask = np.zeros(u.shape, dtype=bool)
    if pos['s2n_st'] == 'True':
        mask1 = validation.sig2noise_val( u, v, s2n, threshold = float(pos['s2n_ra']) )
        mask = mask | mask1
    if pos['gv_st'] == 'True':
        umin, umax = map(float, pos['gv_ul'].split(','))
        vmin, vmax = map(float, pos['gv_vl'].split(','))
        mask2 = validation.global_val( u, v, (umin, umax), (vmin, vmax) )
        mask = mask | mask2
    if pos['lv_st'] == 'True':
        udif, vdif = map(float, pos['lv_df'].split(','))
        mask3 = validation.local_median_val(u, v, udif, vdif, size=int(pos['lv_kr']))
        mask = mask | mask3
    if pos['std_st'] == 'True':
        mask4 = validation.global_std(u, v, std_threshold=float(pos['std_ra']))
        mask = mask | mask4
    # vector corrections
    if pos['bv_st'] == 'True':
        u[mask], v[mask] = np.nan, np.nan
        u, v = filters.replace_outliers( u, v, method=pos['bv_mt'], max_iter=int(pos['bv_ni']), kernel_size=int(pos['bv_kr']))
    if pos['sm_st'] == 'True':
        u_ra, v_ra = map(float, pos['sm_ra'].split(','))
        u, *_ = smoothn.smoothn(u, s=u_ra)
        v, *_ = smoothn.smoothn(v, s=v_ra)
    if pos['fm_st'] == 'True':
        for fm in pos['fm_in'].split(','):
            x, y, u, v, mask = tools.manipulate_field(x, y, u, v, mask, mode=fm.strip())
    # scaling
    scale = float(pro['sc'])
    x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = scale)
    # calculate vorticity and velocity magnitude
    vor = vorticity(u, v, x, y)
    velMag = np.sqrt(u*u + v*v)

    basename = os.path.basename(file_name)
    if saveLoc != None:
        save_name = os.path.join(saveLoc, basename)
        tools.save(x, y, filename=save_name, variables=[u, v, mask, vor, velMag], header=header)

    return basename, u, v, mask, vor, velMag


def simpleOutput(file_list, pro, pos):
    """Manages the post processing for a minimall output and saves the results
    """
    Analysis_folder = os.path.dirname(os.path.dirname(file_list[0]))
    header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude"'
    
    pool = multiprocessing.Pool(processes=int(pro['nc']))
    func = partial(postProcess, pro=pro, pos=pos, saveLoc=Analysis_folder, header=header)
    pool.map(func, file_list)   


def extendedOutput(file_list, pro, pos):
    """Carries out the post processing for an extended output with temporal analysis and saves the results
    """
    
    # initialize variables to hold data
    Nt = len(file_list)
    x, y, u, v, mask = tools.read_data(file_list[0])
    # do field manipulation and scaling on x and y
    if pos['fm_st'] == 'True':
        for fm in pos['fm_in'].split(','):
            x, y, u, v, mask = tools.manipulate_field(x, y, u, v, mask, mode=fm.strip())
    scale = float(pro['sc'])
    x, y = x/scale, y/scale
    u, v, mask, vor, velMag = np.zeros((5, x.shape[0], x.shape[1], Nt), np.float)
    # get all the data
    #for n, fl in enumerate(file_list):
    #    *_, u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n] = postProcess(fl, pro, pos)
    basename = np.zeros((Nt,), 'U50')
    pool = multiprocessing.Pool(processes=int(pro['nc']))
    func = partial(postProcess, pro=pro, pos=pos, saveLoc=None, header=None)
    data = pool.map(func, file_list)
    for n, D in enumerate(data):
        basename[n], u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n] = D
    del(data)

    # calculate mean values
    um, vm, vorm = u.mean(axis=2), v.mean(axis=2), vor.mean(axis=2)
    # calculate fluctuations
    up = u - um[...,np.newaxis]
    vp = v - vm[...,np.newaxis]
    upvp = up * vp
    TKE = 0.5 * (up*up + vp*vp)
    #TKEm = TKE.mean(axis=2) (this is the same as TKE_rms but has more computational cost to calculate)
    upupm = (up*up).mean(axis=2)
    vpvpm = (vp*vp).mean(axis=2)
    upvpm = upvp.mean(axis=2)
    TKE_rms = 0.5 * (upupm + vpvpm)
    # save the results
    Analysis_folder = os.path.dirname(os.path.dirname(file_list[0]))
    avg_file = os.path.join(Analysis_folder, 'AVG.dat')
    header = '"x", "y", "u_avg", "v_avg", "vorticity_avg", "upup_avg", "vpvp_avg", "upvp_avg", "TKE_rms"'
    tools.save( x, y, filename=avg_file, variables=[um, vm, vorm, upupm, vpvpm, upvpm, TKE_rms], header=header )

    header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude", "up", "vp", "upvp", "TKE"'
    for n, bn in enumerate(basename):
        save_name = os.path.join(Analysis_folder, bn)
        tools.save( x, y, filename=save_name, variables=[u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n], up[:,:,n], vp[:,:,n], upvp[:,:,n], TKE[:,:,n]], header=header )


def saveSettings(exp, pre, pro, pos, file_name):
    """Saves the settings given in exp, pre, pro and pos dictionaries to the file_name
    """
    settings = ConfigParser()
    settings['Experiment'] = exp
    settings['Pre-process'] = pre
    settings['Process'] = pro
    settings['Post-process'] = pos
    with open(file_name, 'w') as fl:
        fl.write('# PIV Process Settings:\n\n')
        settings.write(fl)


def loadSettings(file_name):
    """Loads the settings and returns exp, pre, pro and pos dictionaries
    """
    settings = ConfigParser()
    settings.read(file_name)
    
    return settings['Experiment'], settings['Pre-process'], settings['Process'], settings['Post-process']
    

# code to test functions
if __name__ == "__main__":
    r'''
    # vorticity
    filename = r'C:\Users\Asus\Desktop\UI\Dummy Data\fixed000001n.dat'
    x, y, u, v, mask = tools.read_data(filename)
    vor = vorticity(u,v,2.16215,2.16215)
    plt.imshow(vor,cmap='viridis', vmin=-40, vmax=40, interpolation='spline16', origin='lower', extent=(x.min(),x.max(),y.min(),y.max()))
    plt.show()
    
    # AVG
    path = r'E:\Temp\Zoomed in Cylinder\Fixed180\Analysis'
    file_list = glob.glob(path+'\Theta1800deg000???.dat')
    file_list.sort()
    u_avg, v_avg, vor_avg = calcAvgField(file_list, col_list=[2,3,4])
    x, y, u, v, mask = tools.read_data(file_list[0])
    save_file = os.path.join(path, 'AVG.dat')
    tools.save(x, y, u_avg, v_avg, vor_avg, save_file)
    plt.quiver(x,y,u_avg,v_avg)
    plt.show()
    
    # save data
    filename = r'C:\Users\Asus\Desktop\UI\Dummy Data\fixed000001n.dat'
    x, y, u, v, mask = tools.read_data(filename)
    vor = vorticity(u,v,2.16215,2.16215)
    fname = os.path.join(os.path.dirname(filename), 'test.dat')
    tools.save(x, y, filename=fname, header='"x", "y", "u", "v", "mask", "vorticity"', variables=[u, v, mask, vor])
    
    # settings
    filename = r'E:\repos\openpiv-python\openpiv\pouya\UI\Process_Settings.dat'
    exp, pre, pro, pos = loadSettings(filename)
    
    from collections import defaultdict as ddict
    import pprint

    pprint.pprint(pos)
    pos['out_m'] = 'extended'
    saveSettings(exp, pre, pro, pos, filename)
    exp2, pr2e, pro2, pos2 = loadSettings(filename)
    pprint.pprint(pos2)

    
    exp = ddict(lambda: '')
    pre = ddict(lambda: '')
    pro = ddict(lambda: '')
    filename = r'C:\Users\Asus\Desktop\UI\Dummy Data\test.dat'
    saveSettings(exp, pre, pro, pos, filename)
    exp, pre, pro, pos = loadSettings(filename)
    pprint.pprint(exp)
    pprint.pprint(pos)
    '''
    # test output
    setting_file = r'E:\Temp\Processing_Settings.ini'
    exp, pre, pro , pos = OrderedDict(), OrderedDict(), OrderedDict(), OrderedDict()
    exp, pre, pro, pos = loadSettings(setting_file)
    import time
    t1 = time.time()
    for experiment in exp['exp'].split(','):
        for run in exp['run'].split(','):
            experiment, run = experiment.strip(), run.strip()
            path = os.path.join(exp['dir'], experiment, run, 'Analysis', 'Unvalidated')
            file_list = glob.glob(path+'\*.dat')
            file_list.sort()
            Nfiles = int(exp['nf'])
            if Nfiles != 0:
                file_list = file_list[0:Nfiles]
            extendedOutput(file_list, pro, pos)
    print(time.time()-t1)
    '''
    # read data
    datafile = r'E:\Temp\Zoomed in Cylinder\Fixed180\Analysis\Theta1800deg000001.dat'
    x, y, u, v, mask, vor, velMag = tools.read_data(datafile, 7)
    plt.imshow(vor,cmap='viridis', vmin=-100, vmax=100, interpolation='spline16', origin='lower')
    plt.show()
    '''


from PySide2 import QtWidgets
from PySide2.QtGui import QPixmap, Qt, QIcon, QKeySequence
from PySide2.QtCore import QTimer
#import qdarkstyle
import ImageOpener

import matplotlib
#matplotlib.rcParams['figure.facecolor'] = '#19232D'
#matplotlib.rcParams['axes.facecolor'] = '#19232D'
#matplotlib.rc_context({'axes.edgecolor':'xkcd:grey', 'xtick.color':'xkcd:grey', 'ytick.color':'xkcd:grey'})
matplotlib.use('Qt5Agg')
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure

from imageio import imread
import os, time, sys

class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi, tight_layout=True)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class Window(ImageOpener.Ui_MainWindow, QtWidgets.QMainWindow):

    def __init__(self):
        super(Window, self).__init__()
        self.setupUi(self)
        self.setWindowTitle('Image Viewer - Build 001')
        self.setWindowIcon(QIcon("C:\\Users\\Asus\\Desktop\\UI\\ImageOpener\\viewerIcon.png"))
        self.show()
        self.actionOpen.triggered.connect(self.selectFiles)
        self.actionExit.triggered.connect(qApp.exit)
        self.nextPB.clicked.connect(self.nextImage)
        self.previousPB.clicked.connect(self.previousImage)
        # creating key shortcuts
        self.nextShortcut = QtWidgets.QShortcut(QKeySequence('Right'), self)
        self.previousShortcut = QtWidgets.QShortcut(QKeySequence('Left'), self)
        self.nextShortcut.activated.connect(self.nextImage)
        self.previousShortcut.activated.connect(self.previousImage)
        # Create the maptlotlib FigureCanvas object
        self.valplot = MplCanvas(self, width=5, height=4, dpi=100)
        self.valplot.axes.text(5, 5, 'Open images to show from: Files->Open', \
            ha='center', va='center',style='italic', size=14)
        self.valplot.axes.axis([0, 10, 0, 10])
        # Create toolbar, passing canvas as first parent and add to layout
        toolbar = NavigationToolbar(self.valplot, self)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(toolbar)
        layout.addWidget(self.valplot)
        self.frame.setLayout(layout)
        self.files = []
        self.fileN = 0

    def selectFiles(self):
        file_paths, ext = QtWidgets.QFileDialog.getOpenFileNames(self, 'Select Files')
        if file_paths != []:
            self.files = file_paths
            self.fileN = 0
            self.valplot.axes.clear()
            self.viewImage(self.files[0], reset=True)

    def viewImage(self, file, reset=False):
        Img = imread(file)
        xmin, xmax = self.valplot.axes.get_xlim()
        ymin, ymax = self.valplot.axes.get_ylim()
        self.valplot.axes.cla()
        self.valplot.axes.imshow(Img, cmap='gray', vmin=0, vmax=500)
        if reset is False:
            self.valplot.axes.set_xlim(xmin, xmax)
            self.valplot.axes.set_ylim(ymin, ymax)
        self.valplot.draw()
        self.label.setText(os.path.basename(file))

    def nextImage(self):
        if self.fileN < (len(self.files) - 1):
            self.fileN = self.fileN + 1
            self.viewImage(self.files[self.fileN])

    def previousImage(self):
        if self.fileN > 0:
            self.fileN = self.fileN - 1
            self.viewImage(self.files[self.fileN])


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    app.setStyle('Fusion')
    #app.setStyleSheet(qdarkstyle.load_stylesheet_pyside2())
    pixmap = QPixmap("SplashFrog.jpg")
    splash = QtWidgets.QSplashScreen(pixmap)
    splash.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.SplashScreen)
    splash.show()
    QTimer.singleShot(3000, splash.close)
    window = Window()
    sys.exit(app.exec_())





import PIV
from openpiv import tools, validation, filters
import numpy as np
import os
from PySide2 import QtWidgets
import time

import matplotlib
matplotlib.use('Qt5Agg')
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi, tight_layout=True)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)

class Myapp(PIV.Ui_MainWindow, QtWidgets.QMainWindow):
    def __init__(self):
        super(Myapp, self).__init__()
        self.setupUi(self)
        #self.showMaximized()
        self.setWindowTitle('PIV - Build 0.2')
        self.resize(1280, 800)
        self.file_list = {}
        self.first_plot = True
        self.actionload_files.triggered.connect(self.select_files)
        self.actionexit.triggered.connect(qApp.quit)
        self.apply_settings_PB.clicked.connect(lambda: self.update_list(setchange=1))
        self.load_settings_PB.clicked.connect(self.load_settings)
        self.save_settings_PB.clicked.connect(self.save_settings)
        selmodel = self.files_TW.selectionModel()
        selmodel.currentChanged.connect(self.update_plot)
        self.plot_settings_LE.editingFinished.connect(self.update_plot_settings)
        self.BV_settings_CB.currentTextChanged.connect(self.update_plot_settings)
        self.update_plot_settings()

        # Create the maptlotlib FigureCanvas object
        self.valplot = MplCanvas(self, width=5, height=4, dpi=100)
        self.valplot.axes.set_autoscale_on(False)
        self.valplot.axes.text(5, 5, 'Add files to analze using the menubar... \n Then select a file from the list to plot', \
            ha='center', va='center',style='italic', size=12)
        self.valplot.axes.axis([0, 10, 0, 10])
        # Create toolbar, passing canvas as first parament
        toolbar = NavigationToolbar(self.valplot, self)
        layout1 = QtWidgets.QHBoxLayout()
        layout1.addWidget(toolbar)
        layout1.addWidget(self.plot_settings_LE)
        layout1.addWidget(self.BV_settings_CB)
        layout2 = QtWidgets.QVBoxLayout()
        layout2.addLayout(layout1)
        layout2.addWidget(self.valplot)
        # use plot_widget to hold our toolbar and canvas.
        self.plot_widget.setLayout(layout2)

        self.show()

    def select_files(self):
        file_paths, ext = QtWidgets.QFileDialog.getOpenFileNames(self, 'Select Files')
        self.update_list(file_paths=file_paths)
    

    def update_list(self, file_paths=None, setchange=None):
        t1 = time.time()
        if self.files_TW.currentItem():
            current_key = self.files_TW.currentItem().text(0)
        else:
            current_key = None
        self.files_TW.clear()
        if setchange is None:
            for path in file_paths:
                *_, u, v, sig2noise = tools.read_data(path)
                key = os.path.basename(path)
                mask_temp = np.zeros(u.shape, dtype=bool)
                self.file_list[key] = [path, mask_temp, u, v, sig2noise]
                self.file_list[key][1] = self.calc_mask(key=key)
        else:
            for key in self.file_list.keys():
                self.file_list[key][1] = self.calc_mask(key=key)
        for key in self.file_list.keys():
            item = QtWidgets.QTreeWidgetItem(self.files_TW) 
            item.setText(0, key)
            item.setText(1, str(int(np.sum(self.file_list[key][1]))))
            if key == current_key:
                self.files_TW.setCurrentItem(item)
        dt = time.time()-t1
        print(f'processing done in {dt:.2f} sec')

    def calc_mask(self, key):
        u1, v1 = self.file_list[key][2], self.file_list[key][3]
        u, v = u1.copy(), v1.copy()
        if self.s2n_CB.isChecked():
            sig2noise = self.file_list[key][4]
            thr = float(self.s2n_LE.text())
            u, v, mask1 = validation.sig2noise_val( u, v, sig2noise, threshold = thr )
        else:
            mask1 = np.zeros(u.shape, dtype=bool)
        if self.global_velocity_CB.isChecked():
            ulim = [float(i) for i in self.global_uVelocity_LE.text().split(',')]
            vlim = [float(i) for i in self.global_vVelocity_LE.text().split(',')]
            u, v, mask2 = validation.global_val( u, v, (ulim[0], ulim[1]), (vlim[0], vlim[1]) )
        else:
            mask2 = np.zeros(u.shape, dtype=bool)
        if self.local_velocity_CB.isChecked():
            lim = [float(i) for i in self.local_velocity_LE.text().split(',')]
            ksize = int(self.local_kernel_LE.text())
            u, v, mask3 = validation.local_median_val(u, v, lim[0], lim[1], size=ksize)
        else:
            mask3 = np.zeros(u.shape, dtype=bool)
        if self.global_std_CB.isChecked():
            std = float(self.global_std_LE.text())
            u, v, mask4 = validation.global_std(u, v, std_threshold=std)
        else:
            mask4 = np.zeros(u.shape, dtype=bool)

        mask = mask1 | mask2 | mask3 | mask4
        
        return mask

    def update_plot(self, current, previous=None):
        if self.files_TW.itemFromIndex(current):
            t1 = time.time()
            key = self.files_TW.itemFromIndex(current).text(0)
            x, y, u, v, *_ = tools.read_data(self.file_list[key][0])
            mask = self.file_list[key][1]
            valid = ~mask
            if self.first_plot is False:
                xmin, xmax = self.valplot.axes.get_xlim()
                ymin, ymax = self.valplot.axes.get_ylim()
            else:
                xmin, xmax, ymin, ymax = None, None, None, None
                self.first_plot = False

            self.valplot.axes.cla()
            if (self.BV_settings == 'Show Original BV') or (self.BV_settings == 'Only Show BV'):
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Show Replaced BV':
                u[mask], v[mask] = np.nan, np.nan
                u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Do Not Show BV':
                pass
            if self.BV_settings != 'Only Show BV':
                self.valplot.axes.quiver(x[valid], y[valid], u[valid], v[valid], color=self.valid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)

            self.valplot.axes.set_xlim(xmin, xmax)
            self.valplot.axes.set_ylim(ymin, ymax)
            self.valplot.draw()
            dt = time.time()-t1
            print(f'drawing done in {dt:.2f} sec')
    
    def update_plot_settings(self):
        self.BV_settings = self.BV_settings_CB.currentText()
        self.valid_col, self.invalid_col, self.scale, self.width = self.plot_settings_LE.text().split(',')
        self.update_plot(self.files_TW.currentIndex())

    
    def save_settings(self):
        #getting the settings
        s2n_state = str(self.s2n_CB.isChecked())
        s2n_LE = self.s2n_LE.text()
        gv_state = str(self.global_velocity_CB.isChecked())
        gv_LE1 = self.global_uVelocity_LE.text()
        gv_LE2 = self.global_vVelocity_LE.text()
        lv_state = str(self.local_velocity_CB.isChecked())
        lv_LE1 = self.local_velocity_LE.text()
        lv_LE2 = self.local_kernel_LE.text()
        gs_state = str(self.global_std_CB.isChecked())
        gs_LE = self.global_std_LE.text()
        #getting the file path
        path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
            'Select a location to save the settings', 'Validation_Settings.dat')
        #saving to file
        with open(path, 'w') as fh:
            fh.write('Validation settings:\n')
            fh.write(s2n_state+';'+s2n_LE+'\n')
            fh.write(gv_state+';'+gv_LE1+';'+gv_LE2+'\n')
            fh.write(lv_state+';'+lv_LE1+';'+lv_LE2+'\n')
            fh.write(gs_state+';'+gs_LE+'\n')  

    def load_settings(self):
        #load and read file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Validation_Settings.dat')
        lines = []
        with open(path, 'r') as fh:
            for line in fh:
                lines.append(line[:-1])
        #extract and set values
        s2n_state, s2n_LE = lines[1].split(';')
        gv_state, gv_LE1, gv_LE2 = lines[2].split(';')
        lv_state, lv_LE1, lv_LE2 = lines[3].split(';')
        gs_state, gs_LE = lines[4].split(';')
        self.s2n_CB.setChecked(eval(s2n_state))
        self.s2n_LE.setText(s2n_LE)
        self.global_velocity_CB.setChecked(eval(gv_state))
        self.global_uVelocity_LE.setText(gv_LE1)
        self.global_vVelocity_LE.setText(gv_LE2)
        self.local_velocity_CB.setChecked(eval(lv_state))
        self.local_velocity_LE.setText(lv_LE1)
        self.local_kernel_LE.setText(lv_LE2)
        self.global_std_CB.setChecked(eval(gs_state))
        self.global_std_LE.setText(gs_LE)


if __name__ == "__main__":
    app = QtWidgets.QApplication()
    qt_app = Myapp()
    app.exec_()




# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python [conda env:openpiv] *
#     language: python
#     name: conda-env-openpiv-py
# ---

# %%
# # %load red_Cell.py
from openpiv import tools, pyprocess, scaling, filters, \
                    validation, process
import numpy as np
import matplotlib.pyplot as plt

import imageio
from pylab import *
# %matplotlib inline

from skimage import img_as_uint

frame_a  = tools.imread('../test3/Y4-S3_Camera000398.tif')  
frame_b  = tools.imread('../test3/Y4-S3_Camera000399.tif')

# %%
# for whatever reason the shape of frame_a is (3, 284, 256)
# so we first tranpose to the RGB image and then convert to the gray scale

# frame_a = img_as_uint(rgb2gray(frame_a))
# frame_b = img_as_uint(rgb2gray(frame_b))
plt.imshow(np.c_[frame_a,frame_b],cmap=plt.cm.gray)

# %%
# Use Cython version: process.pyx

u, v, sig2noise = process.extended_search_area_piv( frame_a.astype(np.int32), frame_b.astype(np.int32), window_size=32, overlap=8, dt=.1, sig2noise_method='peak2peak' )
x, y = process.get_coordinates( image_size=frame_a.shape, window_size=32, overlap=8 )

u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 1.3 )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 96.52 )

tools.save(x, y, u, v, mask, 'Y4-S3_Camera000398_a.txt' )

# %%
# Use Python version, pyprocess:

u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a.astype(np.int32), frame_b.astype(np.int32), window_size=32, overlap=8, dt=.1, sig2noise_method='peak2peak' )
x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=32, overlap=8 )
u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 1.3 )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 96.52 )

tools.save(x, y, u, v, mask, 'Y4-S3_Camera000398_b.txt' )

# %%
# "natural" view without image
fig,ax = plt.subplots(2,1,figsize=(6,12))
ax[0].invert_yaxis()
ax[0].quiver(x,y,u,v)
ax[0].set_title(' Sort of natural view ')

ax[1].quiver(x,y,u,-v)
ax[1].set_title('Quiver with 0,0 origin needs `negative` v for display');
# plt.quiver(x,y,u,v)

# %%
tools.display_vector_field('Y4-S3_Camera000398_a.txt',on_img=True,image_name='../test3/Y4-S3_Camera000398.tif',scaling_factor=96.52)

# %%
tools.display_vector_field('Y4-S3_Camera000398_a.txt')

# %%
tools.display_vector_field('Y4-S3_Camera000398_b.txt')

# %%
x,y,u,v, mask = process.WiDIM(frame_a.astype(np.int32), frame_b.astype(np.int32), ones_like(frame_a).astype(np.int32), min_window_size=32, overlap_ratio=0.25, coarse_factor=0, dt=0.1, validation_method='mean_velocity', trust_1st_iter=0, validation_iter=0, tolerance=0.7, nb_iter_max=1, sig2noise_method='peak2peak')

# %%
tools.save(x, y, u, v, zeros_like(u), 'Y4-S3_Camera000398_widim1.txt' )

# %%
x,y,u,v, mask = process.WiDIM(frame_a.astype(np.int32), frame_b.astype(np.int32), ones_like(frame_a).astype(np.int32), min_window_size=16, overlap_ratio=0.25, coarse_factor=2, dt=0.1, validation_method='mean_velocity', trust_1st_iter=1, validation_iter=2, tolerance=0.7, nb_iter_max=4, sig2noise_method='peak2peak')

# %%
tools.save(x, y, u, v, zeros_like(u), 'Y4-S3_Camera000398_widim2.txt' )

# %%
tools.display_vector_field('Y4-S3_Camera000398_widim1.txt', widim=True, scale=300, width=0.005)
tools.display_vector_field('Y4-S3_Camera000398_widim2.txt', widim=True, scale=300, width=0.005)
tools.display_vector_field('Y4-S3_Camera000398_a.txt', scale=2, width=0.005,scaling_factor=96.52)
tools.display_vector_field('Y4-S3_Camera000398_b.txt', scale=2, width=0.005,scaling_factor=96.52)



# Fourier Analysis Script
# By: Pouya Mohtat
# Created: Aug. 2020
# Last update: Nov. 2020

# Change Log:
# - added function to perform global short-time fourier analysis which reveals both 'spacial' and 'temporal' variation of frequency fields
# - code refracted and wrapped into a neat analysis object
# - added frequency limit for saved results to keep the output cleaner
# - more efficient data reads (prevent reading the data files multiple times, instead the data is kept in memory)
# - added functions to load and save setting files and logic to run analysis based on this settings

# To do:
# - add ensemble averaging for fft


import numpy as np
import scipy.signal as signal
import os, glob, time
from collections import OrderedDict
from configparser import ConfigParser
import matplotlib.pyplot as plt



class FrequencyAnalysis():
    """class for performing Frequency Analysis. creates the 'Frequency Analysis' folder

    Parameters
    ----------
    path: str
        absolute path to the data files

    nfiles: int
        number of files to use for the analysis

    pattern: str
        the data file naming pattern

    fs: float, optional
        data aquisition frequency, defaults to 1

    dim: float, optional
        scale factor used to nondimentionalize frequency, can be set
        to D/V (D:length scale, V:velocity) to get the Strouhal numb.,
        default is 1 in which case the frequency values are in Hz
    """

    def __init__(self, path, nfiles, pattern, fs=1, dim=1):
        self.N = nfiles
        file_list = glob.glob(os.path.join(path, pattern))
        file_list.sort()
        self.file_list = file_list[:nfiles]
        self.fs = fs
        self.dim = dim
        self.dir = os.path.join(path, 'Frequency Analysis')
        if os.path.isdir(self.dir) == False:
            os.mkdir(self.dir)


    def point_fft(self, u, v):
        """perfomrs FFT for a single point in the flow, saves the results
        and returns frequency and spectral amplitude 
        
        Parameters
        ----------
        u, v : 1D.ndarray
            velocity data at the point

        Returns
        -------
        St : 1d np.ndarray
            Strouhal number or frequency values in Hz

        Su, Sv : 1d np.ndarray
            spectral density values

        """
        Su = abs(np.fft.rfft(u))*2/self.N   #devide by (number of points/2)
        Sv = abs(np.fft.rfft(v))*2/self.N
        St = np.fft.rfftfreq(self.N, 1/self.fs)*self.dim
        Su[0] = Sv[0] = 0    # first values are thrown out
        #saving results in the "Spectral Analysis" directory
        self.saveData(mode='point_fft', St=St, Su=Su, Sv=Sv)
        
        return St, Su, Sv


    def point_stft(self, u, v, nperseg, noverlap):
        """ performs short-time fft for a single point in the flow, saves the results
        and returns frequency, time and spectral amplitudes
        
        Parameters
        ----------
        u, v: 1D.ndarray
            velocity data at the point

        nperseg: int
            number of points in each window

        noverlap: int
            number of points that overlap between each window

        Returns
        -------
        St : 1d np.ndarray
            array containing Strouhal number or frequency in Hz

        t : 1d np.ndarray
            array containing time values

        Su : 1d np.ndarray
            array containing spectral density values

        """
        #run STFT
        f, t, Su = signal.stft(u, fs=self.fs, window='hann', nperseg=nperseg, noverlap=noverlap, \
        return_onesided=True, padded=False)
        f, t, Sv = signal.stft(v, fs=self.fs, window='hann', nperseg=nperseg, noverlap=noverlap, \
        return_onesided=True, padded=False)
        Su[0,:] = Sv[0,:] = 0   #the first value is thrown out
        St = f*self.dim
        Su = abs(Su)
        Sv = abs(Sv)
        #saving results
        self.saveData(mode='point_stft', t=t, St=St, Su=Su, Sv=Sv)
        
        return t, St, Su, Sv


    def global_fft(self, flim, velocity=None):
        """ performs the global Autospectral density (GAS) analysis for velocity data, saves the results
        and returns global specra and the maximum value of GAS at each point
        
        Parameters
        ----------
        flim : float
            the frequecy limit, results for frequencies higher than flim are not saved

        velocity: 2D.ndarray
            velocity array of the shape (nxy, 2). the two columns are u nd v velocity values

        Returns
        -------
        Su_max : 2d np.ndarray
            array of maximum Su values and their corresponding frequency
            at each point

        Su_map : 2d np.ndarray
            array containing the map of Su values at different frequencies, 
            each row represents Su values at a specific frequency

        Sv_max : 2d np.ndarray
            array of maximum Sv values and their corresponding frequency
            at each point

        Sv_map : 2d np.ndarray
            array containing the map of Sv values at different frequencies,
            each row represents Sv values at a specific frequency

        St : 1d np.ndarray
            Strouhal number or Frequency data in Hz

        """
        #initialize arrays
        sample = np.loadtxt(self.file_list[0], skiprows=1)
        nxy = sample.shape[0]
        St = np.fft.rfftfreq(self.N, 1/self.fs)*self.dim
        nfreq = sum(St<flim)                                # number of frequencies that are smaller than the flim
        St = St[0:nfreq]
        Su_max = np.zeros((nxy, 4))                         # maximum Su and its' corresponding frequency at each point
        Su_map = np.zeros((nxy, 2+nfreq))                   # Su values at each frequency (each column corresponds to a certain frequency)
        Su_map[:,0:2] = Su_max[:,0:2] = sample[:,0:2]       # the first two columns are x and y values
        Sv_max = np.zeros((nxy, 4))
        Sv_map = np.zeros((nxy, 2+nfreq))
        Sv_map[:,0:2] = Sv_max[:,0:2] = sample[:,0:2]
        #read velocity data
        if velocity is None:
            velocity = self.getGlobalVelocity()
        #take the fft
        for i in range(nxy):
            Su = abs(np.fft.rfft(velocity[i,0,:])*2/self.N)
            Sv = abs(np.fft.rfft(velocity[i,1,:])*2/self.N)
            Su = Su[0:nfreq]
            Sv = Sv[0:nfreq]
            Su[0] = Sv[0] = 0   #first value gets thrown out
            Su_map[i, 2:] = Su
            Su_max[i, 3] = np.amax(Su)
            Su_max[i, 2] = St[np.argmax(Su)]
            Sv_map[i, 2:] = Sv
            Sv_max[i, 3] = np.amax(Sv)
            Sv_max[i, 2] = St[np.argmax(Sv)]
        #saving the results
        self.saveData(mode='global_fft', Su_max=Su_max, Su_map=Su_map, Sv_max=Sv_max, Sv_map=Sv_map, St=St)
        
        return Su_max, Su_map, Sv_max, Sv_map, St


    def global_stft(self, nperseg, noverlap, flim, velocity=None):
        #read velocity data
        if velocity is None:
            velocity = self.getGlobalVelocity()
        #initialize values
        sample = np.loadtxt(self.file_list[0], skiprows=1)
        nxy = sample.shape[0]
        sample_f, sample_t, sample_Su = signal.stft(velocity[nxy//2], fs=self.fs, window='hann', nperseg=nperseg, noverlap=noverlap, \
            return_onesided=True, padded=False)
        St = sample_f*self.dim
        nfreq = sum(St<flim)                                            # number of frequencies that are smaller than the flim
        St = St[0:nfreq]
        ntime = len(sample_t)
        Su_map = np.zeros((nxy,nfreq+2,ntime), float)                   # Su values at each frequency (each column corresponds to a certain frequency and the third axis is time)
        Sv_map = np.zeros((nxy,nfreq+2,ntime), float)                   # the same...
        Su_map[:,0:2,:] = Sv_map[:,0:2,:] = sample[:,0:2,np.newaxis]    # the first two columns hold values for x and y
        #take the stft
        for i in range(nxy):
            f, t, Su = signal.stft(velocity[i,0,:], fs=self.fs, window='hann', nperseg=nperseg, noverlap=noverlap, \
                return_onesided=True, padded=False)
            f, t, Sv = signal.stft(velocity[i,1,:], fs=self.fs, window='hann', nperseg=nperseg, noverlap=noverlap, \
                return_onesided=True, padded=False)
            Su_map[i,2:,:] = abs(Su[:nfreq,:])
            Sv_map[i,2:,:] = abs(Sv[:nfreq,:])
        #saving the results
        self.saveData(mode='global_stft', t=t, St=St, Su_map=Su_map, Sv_map=Sv_map)

        return t[0:nfreq], St, Su_map, Sv_map


    def getPointVelocity(self, gx, gy, velocity=None):
        self.gx, self.gy = gx, gy
        sample = np.loadtxt(self.file_list[0], skiprows=1)
        index, = np.where((sample[:,0]==self.gx) & (sample[:,1]==self.gy))
        ind = int(index)

        #loop to read files and collect time data
        if velocity is None:
            u = np.zeros(self.N)
            v = np.zeros(self.N)
            for i in range(self.N):
                data = np.loadtxt(self.file_list[i], skiprows=1)
                u[i] = data[ind, 2]
                v[i] = data[ind, 3]
        else:
            u = velocity[ind,0,:]
            v = velocity[ind,1,:]

        return u, v


    def getGlobalVelocity(self):
        sample = np.loadtxt(self.file_list[0], skiprows=1)
        nxy = sample.shape[0]
        velocity = np.zeros((nxy, 2, self.N))
        for i in range(self.N):
            velocity[:,:,i] = np.loadtxt(self.file_list[i], skiprows=1)[:,2:4]
        return velocity


    def saveData(self, mode, t=None, St=None, Su=None, Sv=None, Su_map=None, Su_max=None, Sv_map=None, Sv_max=None):
        if mode == 'point_fft':
            out_u = np.vstack([St, Su])
            out_v = np.vstack([St, Sv])
            headerline = f'TITLE="U_fft for Point=({self.gx},{self.gy})" VARIABLES="St", "Su"'
            np.savetxt(os.path.join(self.dir, f'U_fft_Point ({self.gx},{self.gy}).dat'), out_u.T, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            headerline = f'TITLE="V_fft for Point=({self.gx},{self.gy})" VARIABLES="St", "Sv"'
            np.savetxt(os.path.join(self.dir, f'V_fft_Point ({self.gx},{self.gy}).dat'), out_v.T, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            return

        if mode == 'point_stft':
            t_mesh, St_mesh = np.meshgrid(t, St)
            out_u = np.vstack([t_mesh.ravel(), St_mesh.ravel(), Su.ravel(), Su.ravel()/np.max(Su)])
            out_v = np.vstack([t_mesh.ravel(), St_mesh.ravel(), Sv.ravel(), Sv.ravel()/np.max(Sv)])
            headerline = f'TITLE="U_short_time_fft Point=({self.gx},{self.gy})" VARIABLES="t", "St", "Su", "Su_nondim" ZONE I={t.size}, J={St.size}'
            np.savetxt(os.path.join(self.dir, f'U_stft_Point ({self.gx},{self.gy}).dat'), out_u.T, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            headerline = f'TITLE="V_short_time_fft Point=({self.gx},{self.gy})" VARIABLES="t", "St", "Sv", "Sv_nondim" ZONE I={t.size}, J={St.size}'
            np.savetxt(os.path.join(self.dir, f'V_stft_Point ({self.gx},{self.gy}).dat'), out_v.T, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            return

        sample = np.loadtxt(self.file_list[0], skiprows=1)
        nx = len(np.unique(sample[:,0]))
        ny = sample.shape[0]//nx
        nfreq = Su_map.shape[1] - 2

        if mode == 'global_fft':
            headerline = 'TITLE="Su_map" VARIABLES="x", "y"'
            for n in range(nfreq):
                headerline += (f', "Su for St={St[n]:0.3}"')
            headerline += f' ZONE I={nx}, J={ny}'
            np.savetxt(os.path.join(self.dir, 'Global_Su_map.dat') , Su_map, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            headerline = f'TITLE="Su_max" VARIABLES="x", "y", "St", "Su" ZONE I={nx}, J={ny}'
            np.savetxt(os.path.join(self.dir, 'Global_Su_max.dat'), Su_max, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            headerline = 'TITLE="Sv_map" VARIABLES="x", "y"'
            for n in range(nfreq):
                headerline += (f', "Sv for St={St[n]:0.3}"')
            headerline += f' ZONE I={nx}, J={ny}'
            np.savetxt(os.path.join(self.dir, 'Global_Sv_map.dat') , Sv_map, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            headerline = f'TITLE="Sv_max" VARIABLES="x", "y", "St", "Sv" ZONE I={nx}, J={ny}'
            np.savetxt(os.path.join(self.dir, 'Global_Sv_max.dat'), Sv_max, fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            return

        if mode == 'global_stft':
            for i, t in enumerate(t):
                headerline = f'TITLE="Su_map for time={t:0.3}" VARIABLES="x", "y"'
                for n in range(nfreq):
                    headerline += (f', "Su for St={St[n]:0.3}"')
                headerline += f' ZONE I={nx}, J={ny}'
                np.savetxt(os.path.join(self.dir, f'Global_Su_map_t{t:0.3}.dat') , Su_map[:,:,i], fmt='%8.4f', delimiter='\t', header=headerline, comments='')
                headerline = f'TITLE="Sv_map for time={t:0.3}" VARIABLES="x", "y"'
                for n in range(nfreq):
                    headerline += (f', "Sv for St={St[n]:0.3}"')
                headerline += f' ZONE I={nx}, J={ny}'
                np.savetxt(os.path.join(self.dir, f'Global_Sv_map_t{t:0.3}.dat') , Sv_map[:,:,i], fmt='%8.4f', delimiter='\t', header=headerline, comments='')
            return
    

    @staticmethod
    def saveSettings(exp, analysis, stg_file):
        settings = ConfigParser()
        settings['Experiment'] = exp
        settings['Analysis'] = analysis
        with open(stg_file, 'w') as fl:
            fl.write('# Spectral Analysis Settings:\n\n')
            settings.write(fl)

    @staticmethod
    def loadSettings(stg_file):
        settings = ConfigParser()
        settings.read(stg_file)
        
        return settings['Experiment'], settings['Analysis']
        


#code to test functions
if __name__ == "__main__":
    
    exp, analysis = OrderedDict(), OrderedDict()
    exp, analysis = FrequencyAnalysis.loadSettings('E:\Temp\Re11_medium\Frequency_sttings.ini')
    t1 = time.time()

    experiments = glob.glob(os.path.join(exp['dir'], exp['exp']))
    for experiment in experiments:
        runs = glob.glob(os.path.join(experiment, exp['run']))
        for run in runs:
            path = os.path.join(run, 'Analysis')
            print('initializing...')
            spectral = FrequencyAnalysis(path, int(exp['nf']), exp['pat'], fs=float(analysis['fs']), dim=float(analysis['dim']))
            velocity = None

            if eval(analysis['gb_fft']) or eval(analysis['gb_stft']):
                print('reading velocity...')
                velocity = spectral.getGlobalVelocity()
                if eval(analysis['gb_fft']):
                    print('runing global fft...')
                    spectral.global_fft(flim=float(analysis['flim']), velocity=velocity)
                if eval(analysis['gb_stft']):
                    print('runing global stft...')
                    spectral.global_stft(nperseg=int(analysis['nperseg']), noverlap=int(analysis['noverlap']), flim=float(analysis['flim']), velocity=velocity)
            
            if eval(analysis['pt_fft']) or eval(analysis['pt_stft']):
                print('getting u, v values...')
                if analysis['pt_mode'] == 'specified point':
                    gx, gy = map(float, analysis['pt'].split(','))
                elif analysis['pt_mode'] == 'Max Global Su':
                    fname = os.path.join(path, 'Frequency Analysis', 'Global_Su_max.dat')
                    if os.path.exists(fname):
                        data = np.loadtxt(fname, skiprows=1)
                        amax = np.argmax(data[:,3])
                        gx, gy = data[amax,0], data[amax,1]
                    else:
                        print('No Su data found! skiped, point spectra was not calculated.')
                        gx, gy = None, None       
                elif analysis['pt_mode'] == 'Sv_max':
                    fname = os.path.join(path, 'Frequency Analysis', 'Global_Sv_max.dat')
                    if os.path.exists(fname):
                        data = np.loadtxt(fname, skiprows=1)
                        amax = np.argmax(data[:,3])
                        gx, gy = data[amax,0], data[amax,1]
                    else:
                        print('No Sv data found! skiped, point spectra was not calculated.')
                        gx, gy = None, None
                if gy is not None:
                    u, v = spectral.getPointVelocity(gx, gy, velocity=velocity)
                    if eval(analysis['pt_fft']):
                        print('runing point fft...')
                        spectral.point_fft(u, v)
                    if eval(analysis['pt_stft']):
                        print('runing point stft...')
                        spectral.point_stft(u, v, nperseg=analysis['nperseg'], noverlap=analysis['noverlap'])

    print(f'frequency analysis finished in {time.time()-t1} sec')


# Particle Image Velocimetry GUI
# By Pouya Mohtat Nov. 2020

# Version 0.2.7 change log:
# - spectral analysis tab now active
# - new version format: 0.7 -> 0.2.7
# - the experiment folder addressing is changed and patterns are used instead of single folder names, making the folder addresses more flexible.
# - setting files changed to .ini format for simpler save and load functions
# - 'import raw images' dialogue is removed all together. the same functionality can be achieved using the process tab and then importing unvalidated files.


# to do:
#------------------------------------ 
# 1- stop buttons don't work at the moment. they should provide a gracefull exit from the current running process.
#    it's necessary to have a stop flag of some sort to properly exit from the main 'task.run' multiprocessing loop and have all the child processes 
#    and workers neatly closed and the resources freed up.




from PySide2 import QtWidgets
from PySide2.QtCore import QThread, Signal, QTimer, Qt
from PySide2.QtGui import QIcon, QPixmap
import Main_PIV
from openpivtk import tools, validation, filters, pyprocess, scaling, smoothn, postprocessing, modal, spectral
import numpy as np
import os, sys, glob
from functools import partial
import multiprocessing
from imageio import imsave
from collections import OrderedDict
from configparser import ConfigParser

import matplotlib
matplotlib.use('Qt5Agg')
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi, tight_layout=True)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class MainPIV(Main_PIV.Ui_MainWindow, QtWidgets.QMainWindow):

    def __init__(self):
        super(MainPIV, self).__init__()
        self.setupUi(self)

        #self.showMaximized()
        self.version = '0.2.7'
        self.setWindowTitle(f'PIV Analysis Toolkit V{self.version}')
        self.resize(1280, 800)
        self.show()
        self.file_list = {}
        self.first_plot = True
        self.actionLoad_Files.triggered.connect(self.selectFiles)
        self.run_progress_TE.ensureCursorVisible()
        self.actionExit.triggered.connect(qApp.quit)
        self.actionClear_Files.triggered.connect(self.clearList)
        self.actionAbout.triggered.connect(self.showAbout)
        self.apply_settings_PB.clicked.connect(lambda: self.updateList(setchange=1))
        self.load_settings_PB.clicked.connect(self.loadThrSettings)
        self.save_settings_PB.clicked.connect(self.saveThrSettings)
        selmodel = self.files_TW.selectionModel()
        selmodel.currentChanged.connect(self.updatePlot)
        self.plot_settings_LE.editingFinished.connect(self.updatePlotSettings)
        self.BV_settings_CB.currentTextChanged.connect(self.updatePlotSettings)
        self.updatePlotSettings()
        # Process tab initialization
        self.exp_directory_TB.clicked.connect(lambda: self.getExpDir(tab='process'))
        self.process_savesettings_PB.clicked.connect(self.saveProSettings)
        self.process_loadsettings_PB.clicked.connect(self.loadProSettings)
        self.run_start_PB.clicked.connect(self.startBatchProcessing)
        self.run_stop_PB.clicked.connect(self.stopProcess)
        # Modal tab initialization
        self.mdl_load_from_PB.clicked.connect(self.mdlLoadFrom)
        self.mdl_dir_TB.clicked.connect(lambda: self.getExpDir(tab='modal'))
        self.mdl_load_PB.clicked.connect(self.mdlLoadSettings)
        self.mdl_save_PB.clicked.connect(self.mdlSaveSettings)
        self.mdl_start_PB.clicked.connect(self.mdlStartProcessing)
        self.mdl_stop_PB.clicked.connect(self.stopProcess)
        # Frequency tab initialization
        self.freq_load_from_PB.clicked.connect(self.freqLoadFrom)
        self.freq_dir_TB.clicked.connect(lambda: self.getExpDir(tab='frequency'))
        self.freq_load_PB.clicked.connect(self.freqLoadSettings)
        self.freq_save_PB.clicked.connect(self.freqSaveSettings)
        self.freq_start_PB.clicked.connect(self.freqStartProcessing)
        self.freq_stop_PB.clicked.connect(self.stopProcess)


        # Create the maptlotlib FigureCanvas object
        self.valplot = MplCanvas(self, width=5, height=4, dpi=100)
        self.valplot.axes.text(5, 5, 'Add files to analize using the menubar... \n Then select a file from the list to plot', \
            ha='center', va='center',style='italic', size=12)
        self.valplot.axes.axis([0, 10, 0, 10])
        # Create toolbar, passing canvas as first parent
        toolbar = NavigationToolbar(self.valplot, self)
        layout1 = QtWidgets.QHBoxLayout()
        layout1.addWidget(toolbar)
        layout1.addWidget(self.plot_settings_LE)
        layout1.addWidget(self.BV_settings_CB)
        layout2 = QtWidgets.QVBoxLayout()
        layout2.addLayout(layout1)
        layout2.addWidget(self.valplot)
        # use plot_widget to hold our toolbar and canvas.
        self.plot_widget.setLayout(layout2)

    def selectFiles(self):
        file_paths, ext = QtWidgets.QFileDialog.getOpenFileNames(self, 'Select Files')
        self.updateList(file_paths=file_paths)

    def updateList(self, file_paths=None, setchange=None):
        if self.files_TW.currentItem():
            current_key = self.files_TW.currentItem().text(0)
        else:
            current_key = None
        self.files_TW.clear()
        if setchange is None:
            for path in file_paths:
                *_, u, v, sig2noise = tools.read_data(path)
                key = os.path.basename(path)
                mask_temp = np.zeros(u.shape, dtype=bool)
                self.file_list[key] = [path, mask_temp, u, v, sig2noise]
                self.file_list[key][1] = self.calculateMask(key=key)
        else:
            for key in self.file_list.keys():
                self.file_list[key][1] = self.calculateMask(key=key)
        for key in self.file_list.keys():
            item = QtWidgets.QTreeWidgetItem(self.files_TW) 
            item.setText(0, key)
            item.setText(1, str(int(np.sum(self.file_list[key][1]))))
            if key == current_key:
                self.files_TW.setCurrentItem(item)

    def clearList(self):
        self.files_TW.clear()
        self.file_list = {}
        self.valplot.axes.cla()
        self.valplot.draw()
        self.first_plot = True

    def showAbout(self):
        message = f"PIV Toolkit<br>Version {self.version}<br><br>PIV Toolkit is based on the \
            'OpenPIV' open-source project but with many enhanced capabilities. More information can be found \
            on the projects github page.<br><a href='https://github.com/pouya-m/openpiv-python/tree/PIV-Code-Pouya'>Github page</a><br><br>By Pouya Mohtat"
        msgBox = QtWidgets.QMessageBox.information(self, 'About', message)


    def calculateMask(self, key):
        u, v = self.file_list[key][2], self.file_list[key][3]
        
        if self.s2n_CB.isChecked():
            sig2noise = self.file_list[key][4]
            thr = float(self.s2n_LE.text())
            mask1 = validation.sig2noise_val( u, v, sig2noise, threshold = thr )
        else:
            mask1 = np.zeros(u.shape, dtype=bool)
            
        if self.global_velocity_CB.isChecked():
            ulim = [float(i) for i in self.global_uVelocity_LE.text().split(',')]
            vlim = [float(i) for i in self.global_vVelocity_LE.text().split(',')]
            mask2 = validation.global_val( u, v, (ulim[0], ulim[1]), (vlim[0], vlim[1]) )
        else:
            mask2 = np.zeros(u.shape, dtype=bool)

        if self.local_velocity_CB.isChecked():
            lim = [float(i) for i in self.local_velocity_LE.text().split(',')]
            ksize = int(self.local_kernel_LE.text())
            mask3 = validation.local_median_val(u, v, lim[0], lim[1], size=ksize)
        else:
            mask3 = np.zeros(u.shape, dtype=bool)

        if self.global_std_CB.isChecked():
            std = float(self.global_std_LE.text())
            mask4 = validation.global_std(u, v, std_threshold=std)
        else:
            mask4 = np.zeros(u.shape, dtype=bool)

        mask = mask1 | mask2 | mask3 | mask4
        return mask

    def updatePlot(self, current, previous=None):
        if self.files_TW.itemFromIndex(current):
            key = self.files_TW.itemFromIndex(current).text(0)
            x, y, u, v, *_ = tools.read_data(self.file_list[key][0])
            mask = self.file_list[key][1]
            valid = ~mask
            if self.first_plot is False:
                xmin, xmax = self.valplot.axes.get_xlim()
                ymin, ymax = self.valplot.axes.get_ylim()
            else:
                xmin, xmax, ymin, ymax = None, None, None, None
                self.first_plot = False
            self.valplot.axes.cla()
            if (self.BV_settings == 'Show Original BV') or (self.BV_settings == 'Only Show BV'):
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Show Replaced BV':
                u[mask], v[mask] = np.nan, np.nan
                u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Do Not Show BV':
                pass
            if self.BV_settings != 'Only Show BV':
                self.valplot.axes.quiver(x[valid], y[valid], u[valid], v[valid], color=self.valid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            self.valplot.axes.set_xlim(xmin, xmax)
            self.valplot.axes.set_ylim(ymin, ymax)
            self.valplot.draw()
    
    def updatePlotSettings(self):
        self.BV_settings = self.BV_settings_CB.currentText()
        self.valid_col, self.invalid_col, self.scale, self.width = self.plot_settings_LE.text().split(',')
        self.updatePlot(self.files_TW.currentIndex())

    def saveThrSettings(self):
        #getting the file path
        path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
            'Select a location to save the settings', 'Validation_Settings.ini')
        if path == '':
            return
        #getting the settings
        stg = OrderedDict()
        stg['s2n_state'] = str(self.s2n_CB.isChecked())
        stg['s2n_ratio'] = self.s2n_LE.text()
        stg['gv_state'] = str(self.global_velocity_CB.isChecked())
        stg['gv_ulim'] = self.global_uVelocity_LE.text()
        stg['gv_vlim'] = self.global_vVelocity_LE.text()
        stg['lv_state'] = str(self.local_velocity_CB.isChecked())
        stg['lv_lim'] = self.local_velocity_LE.text()
        stg['lv_kernel'] = self.local_kernel_LE.text()
        stg['std_state'] = str(self.global_std_CB.isChecked())
        stg['std_lim'] = self.global_std_LE.text()
        #saving to file
        settings = ConfigParser()
        settings['Validation'] = stg
        with open(path, 'w') as fh:
            fh.write('# Validation settings:\n\n')
            settings.write(fh)

    def loadThrSettings(self):
        #load and read file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Validation_Settings.ini')
        if path == '':
            return
        settings = ConfigParser()
        settings.read(path)
        #extract and set values
        stg = OrderedDict()
        stg = settings['Validation']
        self.s2n_CB.setChecked(eval(stg['s2n_state']))
        self.s2n_LE.setText(stg['s2n_ratio'])
        self.global_velocity_CB.setChecked(eval(stg['gv_state']))
        self.global_uVelocity_LE.setText(stg['gv_ulim'])
        self.global_vVelocity_LE.setText(stg['gv_vlim'])
        self.local_velocity_CB.setChecked(eval(stg['lv_state']))
        self.local_velocity_LE.setText(stg['lv_lim'])
        self.local_kernel_LE.setText(stg['lv_kernel'])
        self.global_std_CB.setChecked(eval(stg['std_state']))
        self.global_std_LE.setText(stg['std_lim'])
        
    def getExpDir(self, tab):
        dir_path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select Experiment Directory')
        if tab == 'process':
            self.exp_directory_LE.setText(dir_path)
        elif tab == 'modal':
            self.mdl_dir_LE.setText(dir_path)
        elif tab == 'frequency':
            self.freq_dir_LE.setText(dir_path)

    def saveProSettings(self, path=False):
        #getting the file path
        if path is False:
            path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
                'Select a location to save the settings', 'Process_Settings.ini')
        if path == '':
            return
        #getting the settings from GUI
        exp, pre, pro, pos = OrderedDict(), OrderedDict(), OrderedDict(), OrderedDict()
        exp['dir'] = self.exp_directory_LE.text()
        exp['exp'] = self.exp_experiments_LE.text()
        exp['run'] = self.exp_runs_LE.text()
        exp['patA'] = self.exp_patA_LE.text()
        exp['patB'] = self.exp_patB_LE.text()
        exp['nf'] = self.exp_nfiles_LE.text()
        pre['bg_st'] = str(self.pre_background_CB.isChecked())
        pre['bg_nf'] = self.pre_bg_nfiles_LE.text()
        pre['sm_st'] = str(self.pre_staticmask_CB.isChecked())
        pre['sm_pa'] = self.pre_sm_path_LE.text()
        pre['dm_st'] = str(self.pre_dynamicmask_CB.isChecked())
        pro['ws'] = self.pro_windowsize_LE.text()
        pro['sa'] = self.pro_searcharea_LE.text()
        pro['ol'] = self.pro_overlap_LE.text()
        pro['s2n'] = self.pro_sig2noise_CB.currentText()
        pro['ts'] = self.pro_timestep_LE.text()
        pro['sc'] = self.pro_scale_LE.text()
        pro['nc'] = self.pro_ncpus_LE.text()
        pos['s2n_st'] = str(self.pos_sig2noise_CB.isChecked())
        pos['s2n_ra'] = self.pos_s2n_ratio_LE.text()
        pos['gv_st'] = str(self.pos_globalvelocity_CB.isChecked())
        pos['gv_ul'] = self.pos_gv_ulim_LE.text()
        pos['gv_vl'] = self.pos_gv_vlim_LE.text()
        pos['std_st'] = str(self.pos_std_CB.isChecked())
        pos['std_ra'] = self.pos_std_LE.text()
        pos['lv_st'] = str(self.pos_localvelocity_CB.isChecked())
        pos['lv_df'] = self.pos_lv_uvdiff_LE.text()
        pos['lv_kr'] = self.pos_lv_kernel_LE.text()
        pos['bv_st'] = str(self.pos_badvector_CB.isChecked())
        pos['bv_mt'] = self.pos_bv_method_CB.currentText()
        pos['bv_ni'] = self.pos_bv_niterations_LE.text()
        pos['bv_kr'] = self.pos_bv_kernel_LE.text()
        pos['sm_st'] = str(self.pos_smoothing_CB.isChecked())
        pos['sm_ra'] = self.pos_smth_factor_LE.text()
        pos['fm_st'] = str(self.pos_fieldmanip_CB.isChecked())
        pos['fm_in'] = self.pos_fm_LE.text()
        pos['out_m'] = self.pos_output_CB.currentText()
        #saving to file
        postprocessing.saveSettings(exp, pre, pro, pos, path)

        return exp, pre, pro, pos      

    def loadProSettings(self):
        #load setting file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Process_Settings.ini')
        if path == '':
            return
        exp, pre, pro, pos = postprocessing.loadSettings(path)

        self.exp_directory_LE.setText(exp['dir'])
        self.exp_experiments_LE.setText(exp['exp'])
        self.exp_runs_LE.setText(exp['run'])
        self.exp_patA_LE.setText(exp['patA'])
        self.exp_patB_LE.setText(exp['patB'])
        self.exp_nfiles_LE.setText(exp['nf'])
        self.pre_background_CB.setChecked(eval(pre['bg_st']))
        self.pre_bg_nfiles_LE.setText(pre['bg_nf'])
        self.pre_staticmask_CB.setChecked(eval(pre['sm_st']))
        self.pre_sm_path_LE.setText(pre['sm_pa'])
        self.pre_dynamicmask_CB.setChecked(eval(pre['dm_st']))
        self.pro_windowsize_LE.setText(pro['ws'])
        self.pro_searcharea_LE.setText(pro['sa'])
        self.pro_overlap_LE.setText(pro['ol'])
        self.pro_sig2noise_CB.setCurrentText(pro['s2n'])
        self.pro_timestep_LE.setText(pro['ts'])
        self.pro_scale_LE.setText(pro['sc'])
        self.pro_ncpus_LE.setText(pro['nc'])
        self.pos_sig2noise_CB.setChecked(eval(pos['s2n_st']))
        self.pos_s2n_ratio_LE.setText(pos['s2n_ra'])
        self.pos_globalvelocity_CB.setChecked(eval(pos['gv_st']))
        self.pos_gv_ulim_LE.setText(pos['gv_ul'])
        self.pos_gv_vlim_LE.setText(pos['gv_vl'])
        self.pos_std_CB.setChecked(eval(pos['std_st']))
        self.pos_std_LE.setText(pos['std_ra'])
        self.pos_localvelocity_CB.setChecked(eval(pos['lv_st']))
        self.pos_lv_uvdiff_LE.setText(pos['lv_df'])
        self.pos_lv_kernel_LE.setText(pos['lv_kr'])
        self.pos_badvector_CB.setChecked(eval(pos['bv_st']))
        self.pos_bv_method_CB.setCurrentText(pos['bv_mt'])
        self.pos_bv_niterations_LE.setText(pos['bv_ni'])
        self.pos_bv_kernel_LE.setText(pos['bv_kr'])
        self.pos_smoothing_CB.setChecked(eval(pos['sm_st']))
        self.pos_smth_factor_LE.setText(pos['sm_ra'])
        self.pos_fieldmanip_CB.setChecked(eval(pos['fm_st']))
        self.pos_fm_LE.setText(pos['fm_in'])
        self.pos_output_CB.setCurrentText(pos['out_m'])

    
    def startBatchProcessing(self):
        if not os.path.isdir(self.exp_directory_LE.text()):
            QtWidgets.QMessageBox.warning(self, 'Experiment Directory Not Found!', 
                'The selected directory is not a valid path. Please select a valid directory in the "Run Settings" under "Experiment" tab...')
            return
        experiments = glob.glob(os.path.join(self.exp_directory_LE.text(), self.exp_experiments_LE.text()))
        for exp in experiments:
            setting_file = os.path.join(exp, 'Processing_Settings.ini')
            exp, pre, pro, pos = self.saveProSettings(setting_file)
        self.run_start_PB.setEnabled(False)
        manager = multiprocessing.Manager()
        self.processed_files = manager.list()
        self.piv_process_thread = PIVProcessThread(exp, pre, pro, pos, self.processed_files)
        self.piv_process_thread.progress_sig.connect(self.run_progress_TE.appendPlainText)
        self.piv_process_thread.finished.connect(lambda: self.finishProcess(tab='process'))
        self.piv_process_thread.start()
        #initialize the progress bars
        nrun = 0
        experiments =  glob.glob(os.path.join(exp['dir'], exp['exp']))
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, exp['run']))
            nrun += len(runs)
        self.nf = int(exp['nf'])
        ntotal = self.nf*nrun
        self.run_progress_PBar.setEnabled(True)
        self.run_overalprogress_PBar.setEnabled(True)
        self.run_overalprogress_PBar.setRange(0, ntotal)
        self.run_progress_PBar.setRange(0, self.nf)
        self.run_overalprogress_PBar.setValue(0)
        self.run_progress_PBar.setValue(0)
        self.progress_timer = QTimer(self)
        self.progress_timer.timeout.connect(self.updateProgressBar)
        self.progress_timer.start(200)

    def updateProgressBar(self):
        progress = len(self.processed_files)
        self.run_overalprogress_PBar.setValue(progress)
        if progress != 0:
            progress = progress % self.nf
            if progress == 0:
                progress = self.nf
        self.run_progress_PBar.setValue(progress)

    def finishProcess(self, tab):
        if tab == 'process':
            self.run_progress_TE.appendPlainText('All done!')
            self.progress_timer.stop()
            self.run_start_PB.setEnabled(True)
        elif tab == 'modal':
            self.mdl_progress_TE.appendPlainText('All done!')
            self.mdl_start_PB.setEnabled(True)
        elif tab == 'frequency':
            self.freq_progress_TE.appendPlainText('All done!')
            self.freq_start_PB.setEnabled(True)

    def mdlLoadFrom(self):
        self.mdl_dir_LE.setText(self.exp_directory_LE.text())
        self.mdl_exp_LE.setText(self.exp_experiments_LE.text())
        self.mdl_run_LE.setText(self.exp_runs_LE.text())
        self.mdl_nf_LE.setText(self.exp_nfiles_LE.text())

    def mdlLoadSettings(self):
        #load setting file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Modal_Settings.ini')
        if path == '':
            return
        exp, analysis, rec = modal.ModalAnalysis.loadSettings(path)

        self.mdl_dir_LE.setText(exp['dir'])
        self.mdl_exp_LE.setText(exp['exp'])
        self.mdl_pat_LE.setText(exp['pat'])
        self.mdl_run_LE.setText(exp['run'])
        self.mdl_nf_LE.setText(exp['nf'])
        self.mdl_st_CB.setChecked(eval(analysis['st']))
        self.mdl_method_CB.setCurrentText(analysis['mt'])
        self.mdl_nm_LE.setText(analysis['nm'])
        self.mdl_rec_CB.setChecked(eval(rec['st']))
        self.mdl_rec_nm_LE.setText(rec['nm'])
        self.mdl_rec_ns_LE.setText(rec['ns'])

    def mdlSaveSettings(self, path=False):
        #getting the file path
        if path is False:
            path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
                'Select a location to save the settings', 'Modal_Settings.ini')
        if path == '':
            return
        #getting the settings from GUI
        exp, analysis, rec = OrderedDict(), OrderedDict(), OrderedDict()
        exp['dir'] = self.mdl_dir_LE.text()
        exp['exp'] = self.mdl_exp_LE.text()
        exp['pat'] = self.mdl_pat_LE.text()
        exp['run'] = self.mdl_run_LE.text()
        exp['nf'] = self.mdl_nf_LE.text()
        analysis['st'] = str(self.mdl_st_CB.isChecked())
        analysis['mt'] = self.mdl_method_CB.currentText()
        analysis['nm'] = self.mdl_nm_LE.text()
        rec['st'] = str(self.mdl_rec_CB.isChecked())
        rec['nm'] = self.mdl_rec_nm_LE.text()
        rec['ns'] = self.mdl_rec_ns_LE.text()
        #saving to file
        modal.ModalAnalysis.saveSettings(exp, analysis, rec, path)

        return exp, analysis, rec

    def mdlStartProcessing(self):
        # save process settings
        if not os.path.isdir(self.mdl_dir_LE.text()):
            QtWidgets.QMessageBox.warning(self, 'Experiment Directory Not Found!', 
                'The selected directory is not a valid path. Please select a valid directory in the "Experiment" section...')
            return
        experiments = glob.glob(os.path.join(self.mdl_dir_LE.text(), self.mdl_exp_LE.text()))
        for experiment in experiments:
            setting_file = os.path.join(experiment, 'Modal_Settings.ini')
            exp, analysis, rec = self.mdlSaveSettings(setting_file)
        # start the process thread 
        self.mdl_start_PB.setEnabled(False)
        self.mdl_progress_TE.clear()
        self.modal_process_thread = ModalProcessThread(exp, analysis, rec)
        self.modal_process_thread.progresstext_sig.connect(self.mdl_progress_TE.appendPlainText)
        self.modal_process_thread.progressbar_sig.connect(self.mdl_progress_PBar.setValue)
        self.modal_process_thread.finished.connect(lambda: self.finishProcess(tab='modal'))
        self.modal_process_thread.start()
        # set up the progress bar
        nrun = 0
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, exp['run']))
            nrun += len(runs)
        self.mdl_progress_PBar.setEnabled(True)
        self.mdl_progress_PBar.setRange(0, nrun)
        self.mdl_progress_PBar.setValue(0)

    def freqLoadFrom(self):
        self.freq_dir_LE.setText(self.exp_directory_LE.text())
        self.freq_exp_LE.setText(self.exp_experiments_LE.text())
        self.freq_run_LE.setText(self.exp_runs_LE.text())
        self.freq_nf_LE.setText(self.exp_nfiles_LE.text())

    def freqLoadSettings(self):
        #load setting file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Spectral_Settings.ini')
        if path == '':
            return
        exp, analysis = spectral.FrequencyAnalysis.loadSettings(path)

        self.freq_dir_LE.setText(exp['dir'])
        self.freq_exp_LE.setText(exp['exp'])
        self.freq_pat_LE.setText(exp['pat'])
        self.freq_run_LE.setText(exp['run'])
        self.freq_nf_LE.setText(exp['nf'])
        self.freq_fs_LE.setText(analysis['fs'])
        self.freq_dim_LE.setText(analysis['dim'])
        self.freq_flim_LE.setText(analysis['flim'])
        self.freq_pt_fft_CB.setChecked(eval(analysis['pt_fft']))
        self.freq_pt_stft_CB.setChecked(eval(analysis['pt_stft']))
        self.freq_gb_fft_CB.setChecked(eval(analysis['gb_fft']))
        self.freq_gb_stft_CB.setChecked(eval(analysis['gb_stft']))
        self.freq_pt_loc_CB.setCurrentText(analysis['pt_mode'])
        self.freq_pt_loc_LE.setText(analysis['pt'])
        self.freq_nperseg_LE.setText(analysis['nperseg'])
        self.freq_noverlap_LE.setText(analysis['noverlap'])

    def freqSaveSettings(self, path=False):
        #getting the file path
        if path is False:
            path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
                'Select a location to save the settings', 'Spectral_Settings.ini')
        if path == '':
            return
        #getting the settings from GUI
        exp, analysis, rec = OrderedDict(), OrderedDict(), OrderedDict()
        exp['dir'] = self.freq_dir_LE.text()
        exp['exp'] = self.freq_exp_LE.text()
        exp['run'] = self.freq_run_LE.text()
        exp['pat'] = self.freq_pat_LE.text()
        exp['nf'] = self.freq_nf_LE.text()
        analysis['pt_fft'] = str(self.freq_pt_fft_CB.isChecked())
        analysis['pt_stft'] = str(self.freq_pt_stft_CB.isChecked())
        analysis['gb_fft'] = str(self.freq_gb_fft_CB.isChecked())
        analysis['gb_stft'] = str(self.freq_gb_stft_CB.isChecked())
        analysis['fs'] = self.freq_fs_LE.text()
        analysis['dim'] = self.freq_dim_LE.text()
        analysis['flim'] = self.freq_flim_LE.text()
        analysis['pt_mode'] = self.freq_pt_loc_CB.currentText()
        analysis['pt'] = self.freq_pt_loc_LE.text()
        analysis['nperseg'] = self.freq_nperseg_LE.text()
        analysis['noverlap'] = self.freq_noverlap_LE.text()
        #saving to file
        spectral.FrequencyAnalysis.saveSettings(exp, analysis, path)

        return exp, analysis

    def freqStartProcessing(self):
        # save process settings
        if not os.path.isdir(self.freq_dir_LE.text()):
            QtWidgets.QMessageBox.warning(self, 'Experiment Directory Not Found!', 
                'The selected directory is not a valid path. Please select a valid directory in the "Experiment" section...')
            return
        experiments = glob.glob(os.path.join(self.freq_dir_LE.text(), self.freq_exp_LE.text()))
        for experiment in experiments:
            setting_file = os.path.join(experiment, 'Spectral_Settings.ini')
            exp, analysis = self.freqSaveSettings(setting_file)
        # start the process thread 
        self.freq_start_PB.setEnabled(False)
        self.freq_progress_TE.clear()
        self.freq_process_thread = FrequencyProcessThread(exp, analysis)
        self.freq_process_thread.progresstext_sig.connect(self.freq_progress_TE.appendPlainText)
        self.freq_process_thread.progressbar_sig.connect(self.freq_progress_PBar.setValue)
        self.freq_process_thread.finished.connect(lambda: self.finishProcess(tab='frequency'))
        self.freq_process_thread.start()
        # set up the progress bar
        nrun = 0
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, exp['run']))
            nrun += len(runs)
        self.freq_progress_PBar.setEnabled(True)
        self.freq_progress_PBar.setRange(0, nrun)
        self.freq_progress_PBar.setValue(0)

    # still under development, better not use it (directly close the window instead!)
    def stopProcess(self):
        qm = QtWidgets.QMessageBox
        if self.Main_tabs.currentIndex() == 1:
            if hasattr(self, 'piv_process_thread'):
                ans = qm.question(self, "Stop Process", "Are you sure you want to stop the process?", \
                    buttons=qm.StandardButtons(qm.Yes | qm.No), defaultButton=qm.No)
                if ans == qm.No:
                    return
                # this does not work as it's supposed to, the processes are still running because they're busy in 'task.run'
                # loop and they don't check for exitting requests before they are done!
                self.piv_process_thread.quit()
                self.piv_process_thread.stop = True
                self.piv_process_thread.wait() 
                self.run_progress_PBar.setValue(0)
                self.run_overalprogress_PBar.setValue(0)
                self.run_start_PB.setEnabled(True)
                self.progress_timer.stop()
                self.run_progress_TE.appendPlainText('Progress stoped!')
        elif self.Main_tabs.currentIndex() == 3:
            if hasattr(self, 'modal_process_thread'):
                ans = qm.question(self, "Stop Process", "Are you sure you want to stop the process?", \
                    buttons=qm.StandardButtons(qm.Yes | qm.No), defaultButton=qm.No)
                if ans == qm.No:
                    return
                self.modal_process_thread.quit()
                self.mdl_progress_PBar.setValue(0)
                self.mdl_start_PB.setEnabled(True)
                self.mdl_progress_TE.appendPlainText('Progress Stoped!')


class FrequencyProcessThread(QThread):

    def __init__(self, exp, analysis):
        super(FrequencyProcessThread, self).__init__()
        self.exp, self.analysis= exp, analysis
        self.progress = 0

    progresstext_sig = Signal(str)
    progressbar_sig = Signal(int)

    def run(self):
        experiments = glob.glob(os.path.join(self.exp['dir'], self.exp['exp']))
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, self.exp['run']))
            for run in runs:
                path = os.path.join(run, 'Analysis')
                self.progresstext_sig.emit(f'frequency analysis: {run}')
                frequency_analysis = spectral.FrequencyAnalysis(path, int(self.exp['nf']), self.exp['pat'], \
                    fs=float(self.analysis['fs']), dim=float(self.analysis['dim']))
                velocity = None
                self.progresstext_sig.emit('reading velocity data...')
                if eval(self.analysis['gb_fft']) or eval(self.analysis['gb_stft']):
                    velocity = frequency_analysis.getGlobalVelocity()
                    self.progresstext_sig.emit('calculating global spectra...')
                    if eval(self.analysis['gb_fft']):
                        frequency_analysis.global_fft(flim=float(self.analysis['flim']), velocity=velocity)
                    if eval(self.analysis['gb_stft']):
                        frequency_analysis.global_stft(nperseg=int(self.analysis['nperseg']), noverlap=int(self.analysis['noverlap']), \
                            flim=float(self.analysis['flim']), velocity=velocity)
                
                if eval(self.analysis['pt_fft']) or eval(self.analysis['pt_stft']):
                    # to get u, v values first we need gx and gy 
                    if self.analysis['pt_mode'] == 'Specified Point':
                        gx, gy = map(float, self.analysis['pt'].split(','))
                    elif self.analysis['pt_mode'] == 'Max Global Su':
                        fname = os.path.join(path, 'Frequency Analysis', 'Global_Su_max.dat')
                        if os.path.exists(fname):
                            data = np.loadtxt(fname, skiprows=1)
                            amax = np.argmax(data[:,3])
                            gx, gy = data[amax,0], data[amax,1]
                        else:
                            self.progresstext_sig.emit('No Su data found! skiped, point spectra was not calculated.')
                            gx, gy = None, None       
                    elif self.analysis['pt_mode'] == 'Max Global Sv':
                        fname = os.path.join(path, 'Frequency Analysis', 'Global_Sv_max.dat')
                        if os.path.exists(fname):
                            data = np.loadtxt(fname, skiprows=1)
                            amax = np.argmax(data[:,3])
                            gx, gy = data[amax,0], data[amax,1]
                        else:
                            self.progresstext_sig.emit('No Sv data found! skiped, point spectra was not calculated.')
                            gx, gy = None, None
                    if gy is not None:
                        u, v = frequency_analysis.getPointVelocity(gx, gy, velocity=velocity)
                        self.progresstext_sig.emit('calculating point spectra...')
                        if eval(self.analysis['pt_fft']):
                            frequency_analysis.point_fft(u, v)
                        if eval(self.analysis['pt_stft']):
                            frequency_analysis.point_stft(u, v, nperseg=int(self.analysis['nperseg']), noverlap=int(self.analysis['noverlap']))
                self.progress += 1
                self.progressbar_sig.emit(self.progress)


class ModalProcessThread(QThread):

    def __init__(self, exp, analysis, rec):
        super(ModalProcessThread, self).__init__()
        self.exp, self.analysis, self.rec = exp, analysis, rec
        self.progress = 0

    progresstext_sig = Signal(str)
    progressbar_sig = Signal(int)

    def run(self):
        experiments = glob.glob(os.path.join(self.exp['dir'], self.exp['exp']))
        experiments.sort()
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, self.exp['run']))
            runs.sort()
            for run in runs:
                path = os.path.join(run, 'Analysis')
                self.progresstext_sig.emit(f'modal analysis: {run}')
                modal_analysis = modal.ModalAnalysis(path, nfiles=int(self.exp['nf']), pattern=self.exp['pat'])
                if self.analysis['st'] == 'True':
                    if self.analysis['mt'] == 'Singular Value Decomposition':
                        self.progresstext_sig.emit('runing SVD...')
                        modal_analysis.svd(nmode=int(self.analysis['nm']))
                    elif self.analysis['mt'] == 'Snapshots Method':
                        self.progresstext_sig.emit('runing snapshots method...')
                        modal_analysis.snapshot(nmode=int(self.analysis['nm']))
                if self.rec['st'] == 'True':
                    self.progresstext_sig.emit('recontructing flow field...')
                    modal_analysis.reconstructField(nmode=int(self.rec['nm']), nsp=int(self.rec['ns']))
                self.progress += 1
                self.progressbar_sig.emit(self.progress)


class PIVProcessThread(QThread):

    def __init__(self, exp, pre, pro, pos, processed_files):
        super(PIVProcessThread, self).__init__()
        self.exp, self.pre, self.pro, self.pos = exp, pre, pro, pos
        self.processed_files = processed_files
        self.stop = False
        self.stop_timer = QTimer(self)
        self.stop_timer.timeout.connect(self.stopProcess)
        self.stop_timer.start(500)
        
    progress_sig = Signal(str)  #signals have to be defined as class variables

    def run(self):
        experiments = glob.glob(os.path.join(self.exp['dir'], self.exp['exp']))
        experiments.sort()
        for experiment in experiments:
            runs = glob.glob(os.path.join(experiment, self.exp['run']))
            runs.sort()
            for run in runs:
                
                # prepare data directories
                analysis_path = tools.create_directory(run)                #creates the Analysis folder if not already there
                tools.create_directory(analysis_path, folder='Unvalidated')     #creates the Unvalidated folder if not already there
                data_dir = os.path.join(run, 'RawData')
                
                # preprocess
                self.progress_sig.emit(f'Processing run: {run} ')
                task = tools.Multiprocesser( data_dir=data_dir, pattern_a=self.exp['patA'], pattern_b=self.exp['patB'] )
                if self.pre['bg_st'] == 'True':
                    bga_file = os.path.join(analysis_path, 'background_a.TIF')
                    bgb_file = os.path.join(analysis_path, 'background_b.TIF')
                    if ( os.path.exists(bga_file) and os.path.exists(bgb_file) ):
                        self.progress_sig.emit('using existing background files...')
                        background_a = tools.imread(bga_file)
                        background_b = tools.imread(bgb_file)
                    else:
                        self.progress_sig.emit('finding background...')
                        background_a, background_b = task.find_background2(n_files=int(self.pre['bg_nf']))
                        imsave(os.path.join(analysis_path, 'background_a.TIF'), background_a)
                        imsave(os.path.join(analysis_path, 'background_b.TIF'), background_b)
                else:
                    background_a, background_b = None, None
                
                # piv+post process
                self.progress_sig.emit('main process...')
                task.n_files = int(self.exp['nf'])
                Process = partial(mainPIVProcess, bga=background_a, bgb=background_b, pro=self.pro, pos=self.pos, processed_files=self.processed_files)
                data = task.run( func = Process, n_cpus=int(self.pro['nc']) )
                
                # extended output
                if self.pos['out_m'] == 'extended':
                    self.progress_sig.emit('calculating extended output...')
                    # initialize variables to hold data
                    im_file, *_ = glob.glob(os.path.join(data_dir, self.exp['patA']))
                    image = tools.imread(im_file)
                    x, y = pyprocess.get_coordinates(image.shape, int(self.pro['ws']), int(self.pro['ol']))
                    # do field manipulation and scaling on x and y
                    if self.pos['fm_st'] == 'True':
                        for fm in self.pos['fm_in'].split(','):
                            x, y, *_ = tools.manipulate_field(x, y, x, x, x, mode=fm.strip())
                    scale = float(self.pro['sc'])
                    x, y = x/scale, y/scale
                    u, v, mask, vor, velMag = np.zeros((5, x.shape[0], x.shape[1], int(self.exp['nf'])), np.float)
                    basename = np.zeros((int(self.exp['nf']),), 'U50')
                    # extract data
                    for n, D in enumerate(data):
                        basename[n], u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n] = D
                    del(data)
                    # calculate mean values
                    um, vm, vorm = u.mean(axis=2), v.mean(axis=2), vor.mean(axis=2)
                    # calculate fluctuations
                    up = u - um[...,np.newaxis]
                    vp = v - vm[...,np.newaxis]
                    upvp = up * vp
                    TKE = 0.5 * (up*up + vp*vp)
                    upupm = (up*up).mean(axis=2)
                    vpvpm = (vp*vp).mean(axis=2)
                    upvpm = upvp.mean(axis=2)
                    TKE_rms = 0.5 * (upupm + vpvpm)
                    # save the results
                    avg_file = os.path.join(analysis_path, 'AVG.dat')
                    header = '"x", "y", "u_avg", "v_avg", "vorticity_avg", "upup_avg", "vpvp_avg", "upvp_avg", "TKE_rms"'
                    tools.save( x, y, filename=avg_file, variables=[um, vm, vorm, upupm, vpvpm, upvpm, TKE_rms], header=header )

                    header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude", "up", "vp", "upvp", "TKE"'
                    for n, bn in enumerate(basename):
                        fname = os.path.join(analysis_path, bn)
                        tools.save( x, y, filename=fname, variables=[u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n], up[:,:,n], vp[:,:,n], upvp[:,:,n], TKE[:,:,n]], header=header )
                
        return True

    #does not work as intended (timer event is ignored because the program is busy running the 'task.run' function)
    def stopProcess(self):
        if self.stop == True:
            self.quit()
            print('trying to quit...')


def mainPIVProcess( args, bga, bgb, pro, pos, processed_files):
    # unpacking the arguments
    file_a, file_b, counter = args
    # read images
    frame_a  = tools.imread( file_a )
    frame_b  = tools.imread( file_b )
    # background removal
    if bga is not None:
        frame_a = frame_a - bga
        frame_b = frame_b - bgb
    # process image pair with piv algorithm.
    u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a, frame_b, \
        window_size=int(pro['ws']), overlap=int(pro['ol']), dt=float(pro['ts']), search_area_size=int(pro['sa']), sig2noise_method=pro['s2n'])
    x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=int(pro['ws']), overlap=int(pro['ol']) )
    save_path = tools.create_path(file_a, folders=['Analysis', 'Unvalidated'])
    tools.save(x, y, u, v, sig2noise, save_path)
    #post processing
    #validation
    mask = np.zeros(u.shape, dtype=bool)
    if pos['s2n_st'] == 'True':
        mask1 = validation.sig2noise_val( u, v, sig2noise, threshold = float(pos['s2n_ra']) )
        mask = mask | mask1
    if pos['gv_st'] == 'True':
        umin, umax = map(float, pos['gv_ul'].split(','))
        vmin, vmax = map(float, pos['gv_vl'].split(','))
        mask2 = validation.global_val( u, v, (umin, umax), (vmin, vmax) )
        mask = mask | mask2
    if pos['lv_st'] == 'True':
        udif, vdif = map(float, pos['lv_df'].split(','))
        mask3 = validation.local_median_val(u, v, udif, vdif, size=int(pos['lv_kr']))
        mask = mask | mask3
    if pos['std_st'] == 'True':
        mask4 = validation.global_std(u, v, std_threshold=float(pos['std_ra']))
        mask = mask | mask4
    # vector corrections
    if pos['bv_st'] == 'True':
        u[mask], v[mask] = np.nan, np.nan
        u, v = filters.replace_outliers( u, v, method=pos['bv_mt'], max_iter=int(pos['bv_ni']), kernel_size=int(pos['bv_kr']))
    if pos['sm_st'] == 'True':
        u_ra, v_ra = map(float, pos['sm_ra'].split(','))
        u, *_ = smoothn.smoothn(u, s=u_ra)
        v, *_ = smoothn.smoothn(v, s=v_ra)
    if pos['fm_st'] == 'True':
        for fm in pos['fm_in'].split(','):
            x, y, u, v, mask = tools.manipulate_field(x, y, u, v, mask, mode=fm.strip())
    # scaling
    scale = float(pro['sc'])
    x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = scale)
    # calculate vorticity and velocity magnitude
    vor = postprocessing.vorticity(u, v, x, y)
    velMag = np.sqrt(u*u + v*v)
    # update processed files, then save or return values
    processed_files.append(save_path)
    if pos['out_m'] == 'simple':
        header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude"'
        fname = tools.create_path(file_a)
        tools.save(x, y, filename=fname, variables=[u, v, mask, vor, velMag], header=header)
    else:
        basename = os.path.basename(save_path)
        return basename, u, v, mask, vor, velMag

def main():
    app = QtWidgets.QApplication(sys.argv)
    # app.setStyle('Fusion')
    dirname = os.path.dirname(__file__)
    app.setWindowIcon(QIcon(os.path.join(dirname, 'PIV.ico')))
    app.setStyleSheet(open(os.path.join(dirname, 'StyleSheet.qss'), 'r').read())
    # pixmap = QPixmap(os.path.join(dirname, 'Splash.jpg'))
    # splash = QtWidgets.QSplashScreen(pixmap)
    # splash.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.SplashScreen)
    # splash.show()
    # QTimer.singleShot(3000, splash.close)
    main_piv = MainPIV()
    sys.exit(app.exec_())   


if __name__ == "__main__":
    main()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python [conda env:openpiv] *
#     language: python
#     name: conda-env-openpiv-py
# ---

# %% [markdown]
# ## OpenPIV tutorial 2
#
# In this notebook we compare the time to run the same analysis using Cython (precompiled) version
# with the Python process using FFT and/or direct cross-correlation method

# %%
from openpiv import tools, process, scaling, pyprocess, validation, filters
import numpy as np
import pylab
# %matplotlib inline

# %%
frame_a  = tools.imread( '../test1/exp1_001_a.bmp' )
frame_b  = tools.imread( '../test1/exp1_001_b.bmp' )
pylab.imshow(np.c_[frame_a,np.ones((frame_a.shape[0],20)),frame_b],cmap=pylab.cm.gray)

# %%
# %%time
u, v, sig2noise = process.extended_search_area_piv( frame_a.astype(np.int32), frame_b.astype(np.int32), window_size=24, overlap=12, dt=0.02, search_area_size=64, sig2noise_method='peak2peak' )
x, y = process.get_coordinates( image_size=frame_a.shape, window_size=24, overlap=12 )
u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 2.5 )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 96.52 )
tools.save(x, y, u, v, mask, 'exp1_001_extended.txt' )

# %%
# %%time
u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a, frame_b, corr_method='fft', window_size=24, overlap=12, dt=0.02, sig2noise_method='peak2peak' )
x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=24, overlap=12 )
u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 2.5 )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2.5)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 96.52 )
tools.save(x, y, u, v, mask, 'exp1_001_fft.txt' )

# %%
# %%time 
u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a, frame_b, corr_method='direct', window_size=24, overlap=12, dt=0.02, sig2noise_method='peak2peak' )
x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=24, overlap=12 )
u, v, mask = validation.sig2noise_val( u, v, sig2noise, threshold = 2.5 )
u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2.5)
x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = 96.52 )
tools.save(x, y, u, v, mask, 'exp1_001_direct.txt' )

# %%
tools.display_vector_field('exp1_001_extended.txt', scale=30, width=0.0025)
tools.display_vector_field('exp1_001_direct.txt', scale=30, width=0.0025)
tools.display_vector_field('exp1_001_fft.txt', scale=30, width=0.0025);



# Particle Image Velocimetry GUI
# By Pouya Mohtat Nov. 2020

# Version 0.5 change log:
# - progressbars are now updated using Qtimers (a much better implementation)
# - progress TextEdit now updates usings signals and sluts. We cannot set the TextEdit from another thread even if 
#   we pass the TextEdit object to the new thread! Random errors and fatal program exits are now solved
# - ability to calculate mean and fluctuating values (including TKE and Re stress) was added (extended output)
# - changed version convention, added icon and help>about page
# - done some styling on the progressbars and buttons (test styling)

# to do:
#------------------------------------
# 1- local median filter seems to work better with kernal=2 but in some situations it will not catch bad vectors with kernel=2 and we need to specify kernel=1 to catch them.
#    this should be investigated and possibly we should provide an option to apply multiple median filters back to back or with vector replacement in between. 

# 2- there should be an option to use already calculated background image files from previous runs (saved in the analysis folder) instead of calculating them each time.


from PySide2 import QtWidgets
from PySide2.QtCore import QThread, Signal, QTimer, Qt
from PySide2.QtGui import QIcon
import Main_PIV
import ImportDialog_Val as ImportDialog_Val
from openpiv import tools, validation, filters, pyprocess, scaling, smoothn, postprocessing
import numpy as np
import os, sys, glob
from functools import partial
import multiprocessing
from imageio import imsave
import styles

import matplotlib
matplotlib.use('Qt5Agg')
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = Figure(figsize=(width, height), dpi=dpi, tight_layout=True)
        self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(fig)


class ImportDialog(ImportDialog_Val.Ui_ImportDialog_Val, QtWidgets.QDialog):

    def __init__(self):
        super(ImportDialog, self).__init__()
        self.setupUi(self)
        self.folderPath_TB.clicked.connect(self.selectFiles)
        self.addFiles_PB.clicked.connect(self.addFiles)
        self.cancel_PB.clicked.connect(self.close)

    def selectFiles(self):
        dir_path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select Folder Containing Raw Images')
        self.folderPath_LE.setText(dir_path)

    def addFiles(self):
        self.addFiles_PB.setEnabled(False)
        settings = {}
        settings['DP']= self.folderPath_LE.text()
        settings['P_a'] = self.patternA_LE.text()
        settings['P_b'] = self.patternB_LE.text()
        settings['N_f'] = int(self.nFiles_LE.text())
        settings['WS'] = int(self.windowSize_LE.text())
        settings['OL'] = int(self.overlap_LE.text())
        settings['DT'] = float(self.timeStep_LE.text())
        settings['SA'] = int(self.searchSize_LE.text())
        settings['S2N'] = self.sig2noise_CB.currentText()
        settings['N_cpu'] = int(self.ncpus_LE.text())
        manager = multiprocessing.Manager()
        self.processed_files = manager.list()
        self.import_process_thread = ImportProcessThread(settings, self.processed_files)
        self.import_process_thread.start()
        self.import_process_thread.finished.connect(self.finishSimpleProcess)
        
        self.progressBar.setRange(0, settings['N_f'])
        self.progressBar.setValue(0)
        self.progress_timer = QTimer(self)
        self.progress_timer.timeout.connect(self.updateProgressBar)
        self.progress_timer.start(200)

    def updateProgressBar(self):
        self.progressBar.setValue(len(self.processed_files))

    def finishSimpleProcess(self):
        pfile = []
        for pf in self.processed_files:
            base, ext = pf.split('.')
            pfile.append(base + '_unvalidated.' + ext)
        MainPIV.updateList(main_piv, file_paths=pfile)   
        self.close()
    

class ImportProcessThread(QThread):

    def __init__(self, settings, processed_files):
        super(ImportProcessThread, self).__init__()
        self.stg = settings
        self.processed_files = processed_files

    def run(self):
        Run_path = os.path.dirname(self.stg['DP'])
        tools.create_directory(Run_path, folder='Analysis')
        #tools.create_directory(Run_path, folder='Analysis/Validation')
        task = tools.Multiprocesser( data_dir=self.stg['DP'], pattern_a=self.stg['P_a'], pattern_b=self.stg['P_b'] )
        task.n_files = self.stg['N_f']
        process = partial(simpleProcess, processed_files=self.processed_files, WS=self.stg['WS'], OL=self.stg['OL'], \
            DT=self.stg['DT'], SA=self.stg['SA'], S2N=self.stg['S2N'])
        task.run( func = process, n_cpus=self.stg['N_cpu'] )


class MainPIV(Main_PIV.Ui_MainWindow, QtWidgets.QMainWindow):

    def __init__(self):
        super(MainPIV, self).__init__()
        self.setupUi(self)

        #self.showMaximized()
        self.version = 0.4
        self.setWindowTitle(f'Particle Image Velocimetry GUI')
        self.resize(1280, 800)
        self.show()
        self.file_list = {}
        self.first_plot = True
        self.actionLoad_Files.triggered.connect(self.selectFiles)
        self.run_progress_TE.ensureCursorVisible()
        #self.actionExit.triggered.connect(qApp.quit)
        self.actionLoad_Raw_Images.triggered.connect(self.openRawImages)
        self.actionClear_Files.triggered.connect(self.clearList)
        self.actionAbout.triggered.connect(self.showAbout)
        self.apply_settings_PB.clicked.connect(lambda: self.updateList(setchange=1))
        self.load_settings_PB.clicked.connect(self.loadThrSettings)
        self.save_settings_PB.clicked.connect(self.saveThrSettings)
        selmodel = self.files_TW.selectionModel()
        selmodel.currentChanged.connect(self.updatePlot)
        self.plot_settings_LE.editingFinished.connect(self.updatePlotSettings)
        self.BV_settings_CB.currentTextChanged.connect(self.updatePlotSettings)
        self.updatePlotSettings()
        # Process tab initialization
        self.exp_directory_TB.clicked.connect(self.getExpDir)
        self.process_savesettings_PB.clicked.connect(self.saveProSettings)
        self.process_loadsettings_PB.clicked.connect(self.loadProSettings)
        self.run_start_PB.clicked.connect(self.StartBatchProcessing)

        # Create the maptlotlib FigureCanvas object
        self.valplot = MplCanvas(self, width=5, height=4, dpi=100)
        self.valplot.axes.text(5, 5, 'Add files to analize using the menubar... \n Then select a file from the list to plot', \
            ha='center', va='center',style='italic', size=12)
        self.valplot.axes.axis([0, 10, 0, 10])
        # Create toolbar, passing canvas as first parent
        toolbar = NavigationToolbar(self.valplot, self)
        layout1 = QtWidgets.QHBoxLayout()
        layout1.addWidget(toolbar)
        layout1.addWidget(self.plot_settings_LE)
        layout1.addWidget(self.BV_settings_CB)
        layout2 = QtWidgets.QVBoxLayout()
        layout2.addLayout(layout1)
        layout2.addWidget(self.valplot)
        # use plot_widget to hold our toolbar and canvas.
        self.plot_widget.setLayout(layout2)

    def selectFiles(self):
        file_paths, ext = QtWidgets.QFileDialog.getOpenFileNames(self, 'Select Files')
        self.updateList(file_paths=file_paths)

    def updateList(self, file_paths=None, setchange=None):
        if self.files_TW.currentItem():
            current_key = self.files_TW.currentItem().text(0)
        else:
            current_key = None
        self.files_TW.clear()
        if setchange is None:
            for path in file_paths:
                *_, u, v, sig2noise = tools.read_data(path)
                key = os.path.basename(path)
                mask_temp = np.zeros(u.shape, dtype=bool)
                self.file_list[key] = [path, mask_temp, u, v, sig2noise]
                self.file_list[key][1] = self.calculateMask(key=key)
        else:
            for key in self.file_list.keys():
                self.file_list[key][1] = self.calculateMask(key=key)
        for key in self.file_list.keys():
            item = QtWidgets.QTreeWidgetItem(self.files_TW) 
            item.setText(0, key)
            item.setText(1, str(int(np.sum(self.file_list[key][1]))))
            if key == current_key:
                self.files_TW.setCurrentItem(item)

    def clearList(self):
        self.files_TW.clear()
        self.file_list = {}
        self.valplot.axes.cla()
        self.valplot.draw()
        self.first_plot = True

    def showAbout(self):
        message = f"Particle Image Velocimetry<br>Version {self.version}<br><br>Particle image velocimetry \
            GUI based on OpenPIV open-source project with added capabilities.<br><a href='https://github.com\
            /pouya-m/openpiv-python/tree/PIV-Code-Pouya'>Github page</a><br><br>By Pouya Mohtat"
        msgBox = QtWidgets.QMessageBox.information(self, 'About', message)


    def calculateMask(self, key):
        u, v = self.file_list[key][2], self.file_list[key][3]
        
        if self.s2n_CB.isChecked():
            sig2noise = self.file_list[key][4]
            thr = float(self.s2n_LE.text())
            mask1 = validation.sig2noise_val( u, v, sig2noise, threshold = thr )
        else:
            mask1 = np.zeros(u.shape, dtype=bool)
            
        if self.global_velocity_CB.isChecked():
            ulim = [float(i) for i in self.global_uVelocity_LE.text().split(',')]
            vlim = [float(i) for i in self.global_vVelocity_LE.text().split(',')]
            mask2 = validation.global_val( u, v, (ulim[0], ulim[1]), (vlim[0], vlim[1]) )
        else:
            mask2 = np.zeros(u.shape, dtype=bool)

        if self.local_velocity_CB.isChecked():
            lim = [float(i) for i in self.local_velocity_LE.text().split(',')]
            ksize = int(self.local_kernel_LE.text())
            mask3 = validation.local_median_val(u, v, lim[0], lim[1], size=ksize)
        else:
            mask3 = np.zeros(u.shape, dtype=bool)

        if self.global_std_CB.isChecked():
            std = float(self.global_std_LE.text())
            mask4 = validation.global_std(u, v, std_threshold=std)
        else:
            mask4 = np.zeros(u.shape, dtype=bool)

        mask = mask1 | mask2 | mask3 | mask4
        return mask

    def updatePlot(self, current, previous=None):
        if self.files_TW.itemFromIndex(current):
            key = self.files_TW.itemFromIndex(current).text(0)
            x, y, u, v, *_ = tools.read_data(self.file_list[key][0])
            mask = self.file_list[key][1]
            valid = ~mask
            if self.first_plot is False:
                xmin, xmax = self.valplot.axes.get_xlim()
                ymin, ymax = self.valplot.axes.get_ylim()
            else:
                xmin, xmax, ymin, ymax = None, None, None, None
                self.first_plot = False
            self.valplot.axes.cla()
            if (self.BV_settings == 'Show Original BV') or (self.BV_settings == 'Only Show BV'):
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Show Replaced BV':
                u[mask], v[mask] = np.nan, np.nan
                u, v = filters.replace_outliers( u, v, method='localmean', max_iter=10, kernel_size=2)
                self.valplot.axes.quiver(x[mask], y[mask], u[mask], v[mask], color=self.invalid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            elif self.BV_settings == 'Do Not Show BV':
                pass
            if self.BV_settings != 'Only Show BV':
                self.valplot.axes.quiver(x[valid], y[valid], u[valid], v[valid], color=self.valid_col, \
                    units='xy', scale=float(self.scale), width=float(self.width), minlength=0.1, minshaft=1.2)
            self.valplot.axes.set_xlim(xmin, xmax)
            self.valplot.axes.set_ylim(ymin, ymax)
            self.valplot.draw()
    
    def updatePlotSettings(self):
        self.BV_settings = self.BV_settings_CB.currentText()
        self.valid_col, self.invalid_col, self.scale, self.width = self.plot_settings_LE.text().split(',')
        self.updatePlot(self.files_TW.currentIndex())

    def saveThrSettings(self):
        #getting the file path
        path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
            'Select a location to save the settings', 'Thresholding_Settings.dat')
        if path == '':
            return
        #getting the settings
        s2n_state = str(self.s2n_CB.isChecked())
        s2n_LE = self.s2n_LE.text()
        gv_state = str(self.global_velocity_CB.isChecked())
        gv_LE1 = self.global_uVelocity_LE.text()
        gv_LE2 = self.global_vVelocity_LE.text()
        lv_state = str(self.local_velocity_CB.isChecked())
        lv_LE1 = self.local_velocity_LE.text()
        lv_LE2 = self.local_kernel_LE.text()
        gs_state = str(self.global_std_CB.isChecked())
        gs_LE = self.global_std_LE.text()
        #saving to file
        with open(path, 'w') as fh:
            fh.write('Thresholding settings:\n')
            fh.write('{0:<30};{1:<15};{2:<15}\n'.format('Signal 2 Noise:', s2n_state, s2n_LE))
            fh.write('{0:<30};{1:<15};{2:<15};{3:<15}\n'.format('Global Velocity:', gv_state, gv_LE1, gv_LE2))
            fh.write('{0:<30};{1:<15};{2:<15};{3:<15}\n'.format('Local Median:', lv_state, lv_LE1, lv_LE2))
            fh.write('{0:<30};{1:<15};{2:<15}\n'.format('Global Std. Deviation:', gs_state, gs_LE))

    def loadThrSettings(self):
        #load and read file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Thresholding_Settings.dat')
        if path == '':
            return
        lines = []
        with open(path, 'r') as fh:
            for line in fh:
                lines.append(line[:-1])
        #extract and set values
        *_, s2n_state, s2n_LE = [lines[1].split(';')[i].strip() for i in range(3)]
        *_, gv_state, gv_LE1, gv_LE2 = [lines[2].split(';')[i].strip() for i in range(4)]
        *_, lv_state, lv_LE1, lv_LE2 = [lines[3].split(';')[i].strip() for i in range(4)]
        *_, gs_state, gs_LE = [lines[4].split(';')[i].strip() for i in range(3)]
        self.s2n_CB.setChecked(eval(s2n_state))
        self.s2n_LE.setText(s2n_LE)
        self.global_velocity_CB.setChecked(eval(gv_state))
        self.global_uVelocity_LE.setText(gv_LE1)
        self.global_vVelocity_LE.setText(gv_LE2)
        self.local_velocity_CB.setChecked(eval(lv_state))
        self.local_velocity_LE.setText(lv_LE1)
        self.local_kernel_LE.setText(lv_LE2)
        self.global_std_CB.setChecked(eval(gs_state))
        self.global_std_LE.setText(gs_LE)

    def openRawImages(self):
        import_dialog = ImportDialog()
        import_dialog.exec_()
        
    def getExpDir(self):
        dir_path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select Experiment Directory')
        self.exp_directory_LE.setText(dir_path)

    def saveProSettings(self, path=False):
        #getting the file path
        if path is False:
            path, ext = QtWidgets.QFileDialog.getSaveFileName(self, \
                'Select a location to save the settings', 'Process_Settings.dat')
        if path == '':
            return
        #getting the settings from GUI
        exp, pre, pro, pos = {}, {}, {}, {}
        exp['dir'] = self.exp_directory_LE.text()
        exp['exp'] = self.exp_experiments_LE.text()
        exp['run'] = self.exp_runs_LE.text()
        exp['patA'] = self.exp_patA_LE.text()
        exp['patB'] = self.exp_patB_LE.text()
        exp['nf'] = self.exp_nfiles_LE.text()
        pre['bg_st'] = str(self.pre_background_CB.isChecked())
        pre['bg_nf'] = self.pre_bg_nfiles_LE.text()
        pre['bg_cs'] = self.pre_bg_chunksize_LE.text()
        pre['bg_nc'] = self.pre_bg_ncpus_LE.text()
        pre['sm_st'] = str(self.pre_staticmask_CB.isChecked())
        pre['sm_pa'] = self.pre_sm_path_LE.text()
        pre['dm_st'] = str(self.pre_dynamicmask_CB.isChecked())
        pro['ws'] = self.pro_windowsize_LE.text()
        pro['sa'] = self.pro_searcharea_LE.text()
        pro['ol'] = self.pro_overlap_LE.text()
        pro['s2n'] = self.pro_sig2noise_CB.currentText()
        pro['ts'] = self.pro_timestep_LE.text()
        pro['sc'] = self.pro_scale_LE.text()
        pro['nc'] = self.pro_ncpus_LE.text()
        pos['s2n_st'] = str(self.pos_sig2noise_CB.isChecked())
        pos['s2n_ra'] = self.pos_s2n_ratio_LE.text()
        pos['gv_st'] = str(self.pos_globalvelocity_CB.isChecked())
        pos['gv_ul'] = self.pos_gv_ulim_LE.text()
        pos['gv_vl'] = self.pos_gv_vlim_LE.text()
        pos['std_st'] = str(self.pos_std_CB.isChecked())
        pos['std_ra'] = self.pos_std_LE.text()
        pos['lv_st'] = str(self.pos_localvelocity_CB.isChecked())
        pos['lv_df'] = self.pos_lv_uvdiff_LE.text()
        pos['lv_kr'] = self.pos_lv_kernel_LE.text()
        pos['bv_st'] = str(self.pos_badvector_CB.isChecked())
        pos['bv_mt'] = self.pos_bv_method_CB.currentText()
        pos['bv_ni'] = self.pos_bv_niterations_LE.text()
        pos['bv_kr'] = self.pos_bv_kernel_LE.text()
        pos['sm_st'] = str(self.pos_smoothing_CB.isChecked())
        pos['sm_ra'] = self.pos_smth_factor_LE.text()
        pos['fm_st'] = str(self.pos_fieldmanip_CB.isChecked())
        pos['fm_in'] = self.pos_fm_LE.text()
        pos['out_m'] = self.pos_output_CB.currentText()
        #saving to file
        postprocessing.saveSettings(exp, pre, pro, pos, path)

        return exp, pre, pro, pos      

    def loadProSettings(self):
        #load setting file
        path, ext = QtWidgets.QFileDialog.getOpenFileName(self, \
            'Select settings file', 'Process_Settings.dat')
        if path == '':
            return
        exp, pre, pro, pos = postprocessing.loadSettings(path)

        self.exp_directory_LE.setText(exp['dir'])
        self.exp_experiments_LE.setText(exp['exp'])
        self.exp_runs_LE.setText(exp['run'])
        self.exp_patA_LE.setText(exp['patA'])
        self.exp_patB_LE.setText(exp['patB'])
        self.exp_nfiles_LE.setText(exp['nf'])
        self.pre_background_CB.setChecked(eval(pre['bg_st']))
        self.pre_bg_nfiles_LE.setText(pre['bg_nf'])
        self.pre_bg_chunksize_LE.setText(pre['bg_cs'])
        self.pre_bg_ncpus_LE.setText(pre['bg_nc'])
        self.pre_staticmask_CB.setChecked(eval(pre['sm_st']))
        self.pre_sm_path_LE.setText(pre['sm_pa'])
        self.pre_dynamicmask_CB.setChecked(eval(pre['dm_st']))
        self.pro_windowsize_LE.setText(pro['ws'])
        self.pro_searcharea_LE.setText(pro['sa'])
        self.pro_overlap_LE.setText(pro['ol'])
        self.pro_sig2noise_CB.setCurrentText(pro['s2n'])
        self.pro_timestep_LE.setText(pro['ts'])
        self.pro_scale_LE.setText(pro['sc'])
        self.pro_ncpus_LE.setText(pro['nc'])
        self.pos_sig2noise_CB.setChecked(eval(pos['s2n_st']))
        self.pos_s2n_ratio_LE.setText(pos['s2n_ra'])
        self.pos_globalvelocity_CB.setChecked(eval(pos['gv_st']))
        self.pos_gv_ulim_LE.setText(pos['gv_ul'])
        self.pos_gv_vlim_LE.setText(pos['gv_vl'])
        self.pos_std_CB.setChecked(eval(pos['std_st']))
        self.pos_std_LE.setText(pos['std_ra'])
        self.pos_localvelocity_CB.setChecked(eval(pos['lv_st']))
        self.pos_lv_uvdiff_LE.setText(pos['lv_df'])
        self.pos_lv_kernel_LE.setText(pos['lv_kr'])
        self.pos_badvector_CB.setChecked(eval(pos['bv_st']))
        self.pos_bv_method_CB.setCurrentText(pos['bv_mt'])
        self.pos_bv_niterations_LE.setText(pos['bv_ni'])
        self.pos_bv_kernel_LE.setText(pos['bv_kr'])
        self.pos_smoothing_CB.setChecked(eval(pos['sm_st']))
        self.pos_smth_factor_LE.setText(pos['sm_ra'])
        self.pos_fieldmanip_CB.setChecked(eval(pos['fm_st']))
        self.pos_fm_LE.setText(pos['fm_in'])
        self.pos_output_CB.setCurrentText(pos['out_m'])

    
    def StartBatchProcessing(self):
        if os.path.isdir(self.exp_directory_LE.text()):
            directory = os.path.join(self.exp_directory_LE.text(), 'Processing_Settings.dat')
        else:
            QtWidgets.QMessageBox.warning(self, 'Experiment Directory Not Found!', 
                'The selected directory is not a valid path. Please select a valid directory in the "Run Settings" under "Experiment" tab...')
            return
        exp, pre, pro, pos = self.saveProSettings(directory)
        self.run_start_PB.setEnabled(False)
        manager = multiprocessing.Manager()
        self.processed_files = manager.list()
        self.process_thread = ProcessThread(exp, pre, pro, pos, self.processed_files)
        self.process_thread.progress_sig.connect(self.run_progress_TE.appendPlainText)
        self.process_thread.start()
        self.process_thread.finished.connect(self.finishProcess)

        nrun = len(exp['run'].split(','))
        nexp = len(exp['exp'].split(','))
        self.nf = int(exp['nf'])
        ntotal = self.nf*nrun*nexp
        self.run_progress_PBar.setEnabled(True)
        self.run_overalprogress_PBar.setEnabled(True)
        self.run_overalprogress_PBar.setRange(0, ntotal)
        self.run_progress_PBar.setRange(0, self.nf)
        self.run_overalprogress_PBar.setValue(0)
        self.run_progress_PBar.setValue(0)
        self.progress_timer = QTimer(self)
        self.progress_timer.timeout.connect(self.updateProgressBar)
        self.progress_timer.start(200)

    def updateProgressBar(self):
        progress = len(self.processed_files)
        self.run_overalprogress_PBar.setValue(progress)
        if progress != 0:
            progress = progress % self.nf
            if progress == 0:
                progress = self.nf
        self.run_progress_PBar.setValue(progress)

    def finishProcess(self):
        self.run_progress_TE.appendPlainText('All done!')
        self.run_start_PB.setEnabled(True)



class ProcessThread(QThread):

    def __init__(self, exp, pre, pro, pos, processed_files):
        super(ProcessThread, self).__init__()
        self.exp, self.pre, self.pro, self.pos = exp, pre, pro, pos
        self.processed_files = processed_files
        
    progress_sig = Signal(str)  #signals have to be defined as class variables

    def run(self):
        for experiment in self.exp['exp'].split(','):
            for run in self.exp['run'].split(','):
                
                # prepare data directories
                experiment, run = experiment.strip(), run.strip()
                run_path = os.path.join(self.exp['dir'], experiment, run)
                analysis_path = tools.create_directory(run_path)                #creates the Analysis folder if not already there
                tools.create_directory(analysis_path, folder='Unvalidated')     #creates the Unvalidated folder if not already there
                data_dir = os.path.join(run_path, 'RawData')
                
                # preprocess
                self.progress_sig.emit(f'Processing run: {experiment} / {run}')
                task = tools.Multiprocesser( data_dir=data_dir, pattern_a=self.exp['patA'], pattern_b=self.exp['patB'] )
                if self.pre['bg_st'] == 'True':
                    self.progress_sig.emit('finding background...')
                    background_a, background_b = task.find_background(n_files=int(self.pre['bg_nf']), chunk_size=int(self.pre['bg_cs']), n_cpus=int(self.pre['bg_nc']))
                    imsave(os.path.join(analysis_path, 'background_a.TIF'), background_a)
                    imsave(os.path.join(analysis_path, 'background_b.TIF'), background_b)
                else:
                    background_a, background_b = None, None
                
                # piv+post process
                self.progress_sig.emit('main process...')
                task.n_files = int(self.exp['nf'])
                Process = partial(mainProcess, bga=background_a, bgb=background_b, pro=self.pro, pos=self.pos, processed_files=self.processed_files)
                data = task.run( func = Process, n_cpus=int(self.pro['nc']) )
                
                # extended output
                if self.pos['out_m'] == 'extended':
                    self.progress_sig.emit('calculating extended output...')
                    # initialize variables to hold data
                    im_file, *_ = glob.glob(os.path.join(data_dir, self.exp['patA']))
                    image = tools.imread(im_file)
                    x, y = pyprocess.get_coordinates(image.shape, int(self.pro['ws']), int(self.pro['ol']))
                    # do field manipulation and scaling on x and y
                    if self.pos['fm_st'] == 'True':
                        for fm in self.pos['fm_in'].split(','):
                            x, y, *_ = tools.manipulate_field(x, y, x, x, x, mode=fm.strip())
                    scale = float(self.pro['sc'])
                    x, y = x/scale, y/scale
                    u, v, mask, vor, velMag = np.zeros((5, x.shape[0], x.shape[1], int(self.exp['nf'])), np.float)
                    basename = np.zeros((int(self.exp['nf']),), 'U50')
                    # extract data
                    for n, D in enumerate(data):
                        basename[n], u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n] = D
                    del(data)
                    # calculate mean values
                    um, vm, vorm = u.mean(axis=2), v.mean(axis=2), vor.mean(axis=2)
                    # calculate fluctuations
                    up = u - um[...,np.newaxis]
                    vp = v - vm[...,np.newaxis]
                    upvp = up * vp
                    TKE = 0.5 * (up*up + vp*vp)
                    upupm = (up*up).mean(axis=2)
                    vpvpm = (vp*vp).mean(axis=2)
                    upvpm = upvp.mean(axis=2)
                    TKE_rms = 0.5 * (upupm + vpvpm)
                    # save the results
                    avg_file = os.path.join(analysis_path, 'AVG.dat')
                    header = '"x", "y", "u_avg", "v_avg", "vorticity_avg", "upup_avg", "vpvp_avg", "upvp_avg", "TKE_rms"'
                    tools.save( x, y, filename=avg_file, variables=[um, vm, vorm, upupm, vpvpm, upvpm, TKE_rms], header=header )

                    header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude", "up", "vp", "upvp", "TKE"'
                    for n, bn in enumerate(basename):
                        fname = os.path.join(analysis_path, bn)
                        tools.save( x, y, filename=fname, variables=[u[:,:,n], v[:,:,n], mask[:,:,n], vor[:,:,n], velMag[:,:,n], up[:,:,n], vp[:,:,n], upvp[:,:,n], TKE[:,:,n]], header=header )
                
        return True


def mainProcess( args, bga, bgb, pro, pos, processed_files):
    # unpacking the arguments
    file_a, file_b, counter = args
    # read images
    frame_a  = tools.imread( file_a )
    frame_b  = tools.imread( file_b )
    # background removal
    if bga is not None:
        frame_a = frame_a - bga
        frame_b = frame_b - bgb
    # process image pair with piv algorithm.
    u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a, frame_b, \
        window_size=int(pro['ws']), overlap=int(pro['ol']), dt=float(pro['ts']), search_area_size=int(pro['sa']), sig2noise_method=pro['s2n'])
    x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=int(pro['ws']), overlap=int(pro['ol']) )
    save_path = tools.create_path(file_a, folders=['Analysis', 'Unvalidated'])
    tools.save(x, y, u, v, sig2noise, save_path)
    #post processing
    #validation
    mask = np.zeros(u.shape, dtype=bool)
    if pos['s2n_st'] == 'True':
        mask1 = validation.sig2noise_val( u, v, sig2noise, threshold = float(pos['s2n_ra']) )
        mask = mask | mask1
    if pos['gv_st'] == 'True':
        umin, umax = map(float, pos['gv_ul'].split(','))
        vmin, vmax = map(float, pos['gv_vl'].split(','))
        mask2 = validation.global_val( u, v, (umin, umax), (vmin, vmax) )
        mask = mask | mask2
    if pos['lv_st'] == 'True':
        udif, vdif = map(float, pos['lv_df'].split(','))
        mask3 = validation.local_median_val(u, v, udif, vdif, size=int(pos['lv_kr']))
        mask = mask | mask3
    if pos['std_st'] == 'True':
        mask4 = validation.global_std(u, v, std_threshold=float(pos['std_ra']))
        mask = mask | mask4
    # vector corrections
    if pos['bv_st'] == 'True':
        u[mask], v[mask] = np.nan, np.nan
        u, v = filters.replace_outliers( u, v, method=pos['bv_mt'], max_iter=int(pos['bv_ni']), kernel_size=int(pos['bv_kr']))
    if pos['sm_st'] == 'True':
        u_ra, v_ra = map(float, pos['sm_ra'].split(','))
        u, *_ = smoothn.smoothn(u, s=u_ra)
        v, *_ = smoothn.smoothn(v, s=v_ra)
    if pos['fm_st'] == 'True':
        for fm in pos['fm_in'].split(','):
            x, y, u, v, mask = tools.manipulate_field(x, y, u, v, mask, mode=fm.strip())
    # scaling
    scale = float(pro['sc'])
    x, y, u, v = scaling.uniform(x, y, u, v, scaling_factor = scale)
    # calculate vorticity and velocity magnitude
    vor = postprocessing.vorticity(u, v, x, y)
    velMag = np.sqrt(u*u + v*v)
    # update processed files, then save or return values
    processed_files.append(save_path)
    if pos['out_m'] == 'simple':
        header = '"x", "y", "u", "v", "mask", "vorticity", "velocity magnitude"'
        fname = tools.create_path(file_a)
        tools.save(x, y, filename=fname, variables=[u, v, mask, vor, velMag], header=header)
    else:
        basename = os.path.basename(save_path)
        return basename, u, v, mask, vor, velMag

       

def simpleProcess( args, processed_files, WS, OL, DT, SA, S2N):
    # unpacking the arguments
    file_a, file_b, counter = args
    # read images into numpy arrays
    frame_a  = tools.imread( file_a )
    frame_b  = tools.imread( file_b )
    # process image pair with piv algorithm.
    u, v, sig2noise = pyprocess.extended_search_area_piv( frame_a, frame_b, \
        window_size=WS, overlap=OL, dt=DT, search_area_size=SA, sig2noise_method=S2N)
    x, y = pyprocess.get_coordinates( image_size=frame_a.shape, window_size=WS, overlap=OL )
    save_path = tools.create_path(file_a, folder='Analysis')
    tools.save(x, y, u, v, sig2noise, save_path+'_unvalidated.dat')
    processed_files.append(save_path+'.dat')


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    #app.setStyle('Fusion')
    dirname = os.path.dirname(__file__)
    app.setWindowIcon(QIcon(os.path.join(dirname, 'PIV.ico')))
    app.setStyleSheet(open(os.path.join(dirname, 'StyleSheet.qss'), 'r').read())
    #styles.dark(app)
    main_piv = MainPIV()
    sys.exit(app.exec_())
